#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
from os import path, listdir
from matplotlib import pyplot, cm, colors, rcParams
from screenshot_utility import Screenshot
from hemisphere_plot_utility import OrthogonalProjection, plot_hemisphere, plot_annotated_point, \
    plot_annotated_length, plot_right_triangle, get_euler_projection
from line_sampling_ported import prepare_solid_angle_line_sampling
from imageio import imread, imwrite
import subprocess
import re

# Force TrueType fonts to make CGF happy
rcParams["pdf.fonttype"] = 42
rcParams["ps.fonttype"] = 42

# The width of a column in the CGF layout in centimeters
column_width = 8.4


def get_figure_size(width_in_centimeters, height_in_centimeters=None, scaling=1.8):
    """
    Returns a figure size as expected by matplotlib.pyplot.figure. It is
    computed from the given size in centimeters. If no height or width is
    given, the aspect ratio is fixed to be 4:3. The figure size is scaled up by
    the given factor.
    """
    if not height_in_centimeters:
        height_in_centimeters = width_in_centimeters*0.75
    if not width_in_centimeters:
        width_in_centimeters = height_in_centimeters/0.75
    return width_in_centimeters * scaling / 2.54, height_in_centimeters * scaling / 2.54


# The color used for hemisphere lines inside the solid angle of the light
# source (before it is scaled)
solid_angle_line_color = np.asarray([0.95, 0.8, 0.4])
# The color used to shade solid angles that correspond to a light source
solid_angle_color = np.asarray([1.0, 0.9, 0.5])
# The color of the light source
light_color = np.asarray([1.0, 0.995, 0.85])


def list_experiments(file_name_re, experiment_path="../vulkan_renderer/data/experiments"):
    """
    Goes through the given directory and lists files matching a given regular
    expression.
    :param file_name_re: A regular expression for file names.
    :param experiment_path: The directory that is searched.
    :return: A list of pairs (file_path, name). file_path is the complete path
        to a file with matching file name in the given subdirectory (not
        recursively). name is the file name with timing information between an
        underscore and the file extension removed.
    """
    return [(path.join(experiment_path, file_name), re.sub(r"^(.*)_[0-9.]+(\.\w+)$", r"\1\2", file_name))
            for file_name in listdir(experiment_path)
            if re.match(file_name_re, file_name) is not None]


def print_experiment_rmses(file_name_re, reference_name, experiment_path="../vulkan_renderer/data/experiments"):
    """
    Lists experiments through list_experiments(). The specified files should be
    *.hdr. It then computes RMSEs with respect to the experiment with the given
    name (as returned by list_experiments()) and prints them.
    """
    experiments = list_experiments(file_name_re, experiment_path)
    experiments = sorted(experiments)
    references = [path for path, name in experiments if name == reference_name]
    if len(references) != 1:
        print("There are %d references of name %s among experiments matching %s. Need exactly one." % (len(references), reference_name, file_name_re))
        return
    reference = imread(references[0])
    print("Printing RMSEs with respect to %s:" % reference_name)
    mean_factor = 1.0 / np.sqrt(reference.shape[0] * reference.shape[1])
    for path, name in experiments:
        print(name, end=":\t")
        image = imread(path)
        rmse = np.linalg.norm(image - reference) * mean_factor
        print("%.3f" % rmse)


def get_closest_point(line_begin, line_end):
    """
    Returns the normalized direction towards the closest point to the origin on
    a line through the given two points.
    """
    line_direction = line_end - line_begin
    line_direction /= np.linalg.norm(line_direction)
    closest_point = line_begin - (line_begin @ line_direction) * line_direction
    closest_point /= np.linalg.norm(closest_point)
    return closest_point


def magnify_teaser():
    """
    Creates magnified insets for the screenshots of the teaser.
    """
    for file_path, name in list_experiments("bistro_outside_.*.png"):
        screenshot = Screenshot(file_path)
        magnifier_list = [
            (1, (1610, 667), "C2"),
            (2, (1382, 920), "C3"),
            (0, (1546, 102), "C1"),
        ]
        padding = 48
        zoom = 6
        size = (64, 48)
        for position, left_top, color in magnifier_list:
            inset_left_top = (padding, padding + (size[1] * zoom + padding) * position)
            screenshot.add_magnifier(left_top, size, inset_left_top, zoom, color, 15)
        screenshot.save(path.join("../Graphics/line_teaser", name), name == "bistro_outside_reference.png")


def plot_axis(projection, origin, direction, label, other_side=False, label_offset=0.07):
    """
    Plots the coordinate axis with the given direction and label using the
    given world space origin. The label can be flipped to the other side if it
    looks bad. The plot may be skipped if the axis projects to depth.
    """
    origin = np.asarray(origin)
    transformed_origin = projection.transform(origin)
    transformed_direction = projection.transform(direction)
    if np.linalg.norm(transformed_direction[0:2]) > 1.0e-3:
        pyplot.arrow(transformed_origin[0], transformed_origin[1], transformed_direction[0], transformed_direction[1],
                     head_width=0.08, facecolor="w")
        rotation = np.asarray([[0.0, -1.0], [1.0,0.0]])
        text_position = transformed_origin[0:2] + 0.5 * transformed_direction[0:2]
        text_position += (-1.0 if other_side else 1.0) * label_offset * rotation @ transformed_direction[0:2]
        pyplot.text(text_position[0], text_position[1], label, horizontalalignment="center", verticalalignment="center")


def plot_line_solid_angle_single(name, line, line_begin, line_end, radius,
                                 line_direction, line_normal, opening_angle, projection):
    """
    Visualizes a cylinder light in relation to a shading point and how its solid
    angle can be computed.
    """
    closest_point = get_closest_point(line_begin, line_end)
    # Plot a hemisphere
    figure = pyplot.figure(figsize=get_figure_size(column_width, 0.7 * column_width))
    axes = pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False)

    def in_solid_angle(position):
        s = np.einsum("i, ...i -> ...", line_direction, position)
        t = np.einsum("i, ...i -> ...", closest_point, position)
        u = np.einsum("i, ...i -> ...", line_normal, position)
        angle = np.arctan2(u, t)
        return (s > line.begin_s) & (s < line.begin_s + line.extent_s) & (np.abs(angle) < opening_angle)

    def covered_by_solid_angle(position):
        depth = projection.transform(position)[..., 2]
        row = projection.matrix[2, :]
        reverse_position = position - 2.0 * depth[..., np.newaxis] * row.reshape([1] * (position.ndim - 1) + [3])
        return in_solid_angle(position) | in_solid_angle(reverse_position)
    plot_hemisphere(projection, 8, 4, covered_by_solid_angle, color=solid_angle_line_color)
    plot_hemisphere(projection, 8, 4, lambda position: ~covered_by_solid_angle(position))
    # Plot the cylinder
    phase = np.linspace(-np.pi, np.pi, 513)
    boundary_offset_direction = np.cross(line_direction, projection.matrix[2, :])
    boundary_offset_direction /= np.linalg.norm(boundary_offset_direction)
    boundary_offset_normal = np.cross(boundary_offset_direction, line_direction)
    bottom_line = list()
    top_line = list()
    cap_list = list()
    for normal_sign, cap_center in [(-1.0, line_begin), (1.0, line_end)]:
        cap = (cap_center[np.newaxis, :]
               + line_normal * np.cos(phase)[:, np.newaxis] * radius
               + closest_point[np.newaxis, :] * np.sin(phase)[:, np.newaxis] * radius)
        cap_list.append(cap)
        # It may be necessary to plot only half the cap
        if np.dot(line_direction * normal_sign, projection.matrix[2, :]) > 0.0:
            projection.plot_curve(cap, color="k")
        else:
            halves = [(-1.0, 0.85 * light_color, "--", -300.0), (1.0, "k", "-", 0.0)]
            for half_sign, half_color, half_style, half_order in halves:
                half_cap = (cap_center[np.newaxis, :]
                       + half_sign * boundary_offset_normal * np.cos(0.5 * phase)[:, np.newaxis] * radius
                       + boundary_offset_direction * np.sin(0.5 * phase)[:, np.newaxis] * radius)
                projection.plot_curve(half_cap, color=half_color, linestyle=half_style, zorder=half_order)
        bottom_line.append(cap_center - boundary_offset_direction * radius)
        top_line.append(cap_center + boundary_offset_direction * radius)
    projection.plot_curve(np.asarray(bottom_line), color="k")
    projection.plot_curve(np.asarray(top_line), color="k")
    # Shade the cylinder
    for cap in cap_list:
        cap_boundary = projection.transform(cap)
        pyplot.fill(cap_boundary[:, 0], cap_boundary[:, 1], color=light_color, zorder=-3000.0)
    rectangle = np.asarray(bottom_line + top_line[::-1])
    rectangle_boundary = projection.transform(rectangle)
    pyplot.fill(rectangle_boundary[:, 0], rectangle_boundary[:, 1], color=light_color, zorder=-3000.0)
    # Annotate a few points
    origin = projection.transform(np.zeros(3))
    plot_annotated_point(origin[0], origin[1], r"$0$", "left", "top", [0.02, 0.0])
    projection.plot_point_set(np.zeros(3))
    min_distance = 1.0 / line.rcp_min_distance
    closest_point_on_line = closest_point * min_distance
    point_list = [(closest_point_on_line, r"$q$"),
                  (line_begin, r"$l_0$"), (line_end, r"$l_1$")]
    if name == "ut":
        point_list = point_list[0:1]
    for point, label in point_list:
        point_x, point_y = projection.transform(point)[0:2]
        plot_annotated_point(point_x, point_y, label, "left", "bottom", [0.017, 0.0])
    # Annotate the radius
    if name == "perspective":
        radius_line = np.asarray([line_begin, line_begin + boundary_offset_direction * radius])
        projection.plot_curve(radius_line, color="k")
        center_x, center_y = np.mean(projection.transform(radius_line)[:, 0:2], axis=0)
        pyplot.text(center_x + 0.03, center_y, r"$R$")
    if name == "st":
        plot_annotated_length(True, min_distance - radius, min_distance + radius, 0.83, r"$2R$")
    # Annotate the opening angle
    if name == "ut":
        silhouette_point = np.cos(opening_angle) * closest_point + np.sin(opening_angle) * line_normal
        silhouette_point *= np.sqrt(min_distance ** 2 - radius ** 2)
        triangle = [closest_point_on_line, silhouette_point, np.zeros(3)]
        plot_right_triangle(projection, triangle, r"", r"$R$", "", 0.55, r"$\varphi_R$", 0.1, (0.09, 0.09, 0.09))
    # Annotate the extent along the s-axis
    if name == "st":
        for line_point in [line_begin, line_end]:
            to_line = np.asarray([np.zeros(3), line_point])
            projection.plot_curve(to_line, color="k", linestyle=":")
            plot_annotated_length(False, line.begin_s, line.begin_s + line.extent_s, 1.015,
                                  "", r"$l_{0,s}$", r"$l_{1,s}$", dict(), 0.06)
    # Plot the construction of the closest point
    if name == "perspective":
        central_axis = np.asarray([line_begin, line_end])
        projection.plot_curve(central_axis, linestyle=":", color=0.8 * light_color)
        triangle = np.asarray([np.zeros(3), closest_point_on_line, line_end])
        plot_right_triangle(projection, triangle, "", "", "")
    # Plot the coordinate frame constructed for the cylinder
    if name=="perspective":
        axes_origin = (-1.0, -0.33, 0.5)
    elif name=="ut":
        axes_origin = (0.0, 1.07, -0.15)
    elif name=="st":
        axes_origin = (-1.0, -2.4, 0.02)
    axes_length = 1.0
    plot_axis(projection, axes_origin, line_direction * axes_length, r"$s$")
    plot_axis(projection, axes_origin, closest_point * axes_length, r"$t$", True)
    plot_axis(projection, axes_origin, line_normal * axes_length, r"$u$", name == "perspective",
              0.2 if name == "perspective" else 0.07)
    # Plot the solid angle of the cylinder
    side_list = list()
    for sign in [-1, 1]:
        s = np.linspace(line.begin_s, line.begin_s + line.extent_s, 513)
        tu = np.sqrt(1.0 - s * s)
        t = np.cos(opening_angle) * tu
        u = sign * np.sin(opening_angle) * tu
        side = (s[:, np.newaxis] * line_direction[np.newaxis, :]
                + t[:, np.newaxis] * closest_point[np.newaxis, :]
                + u[:, np.newaxis] * line_normal[np.newaxis, :])
        side_list.append(side)
    cap_list = list()
    for s in [line.begin_s, line.begin_s + line.extent_s]:
        angle = np.linspace(-opening_angle, opening_angle, 513)
        radius_factor = np.sqrt(1.0 - s * s)
        cap = (s * line_direction[np.newaxis, :]
               + np.cos(angle)[:, np.newaxis] * closest_point[np.newaxis, :] * radius_factor
               + np.sin(angle)[:, np.newaxis] * line_normal[np.newaxis, :] * radius_factor)
        cap_list.append(cap)
    boundary_list = [side_list[0], cap_list[1], side_list[1][::-1], cap_list[0][::-1]]
    full_boundary = np.vstack(boundary_list)
    projection.plot_curve(full_boundary, color=solid_angle_line_color)
    # Remember the limits so that an image plot will not ruin them
    x_lim = axes.get_xlim()
    y_lim = axes.get_ylim()
    # Shade the solid angle using rasterization
    x_min, x_max = (-1.0, 1.0)
    y_min, y_max = (y_lim[0], 1.0)
    width, height = (1024, 1024)
    x, y = np.meshgrid(np.linspace(x_min, x_max, width), np.linspace(y_min, y_max, height))
    z = np.sqrt(1.0 - x * x - y * y)
    view_space = np.zeros(list(x.shape) + [3])
    view_space[:, :, 0] = x
    view_space[:, :, 1] = y
    view_space[:, :, 2] = z
    world_space = np.einsum("ij, ...i-> ...j", projection.matrix, view_space)
    shade = covered_by_solid_angle(world_space)
    image = np.ones(list(x.shape) + [3])
    for j in range(3):
        image[..., j][shade] = solid_angle_color[j]
    pyplot.imshow(image, extent=(x_min, x_max, y_min, y_max), zorder=-1000.0, origin="lower")
    if name=="ut":
        axes.set_xlim(x_lim[0] - 0.09, x_lim[1])
        axes.set_ylim(y_lim[0] - 0.06, y_lim[1])
    else:
        axes.set_xlim(x_lim)
        axes.set_ylim(y_lim)
    figure.savefig("../Graphics/line_solid_angle/%s.pdf" % name, bbox_inches="tight", pad_inches=-0.12)


def plot_line_solid_angle():
    """
    Invokes plot_line_solid_angle_single() with all appropriate perspectives.
    """
    # Prepare a cylinder
    line_begin = np.asarray([-1.3, -0.3, 1.8])
    line_end = np.asarray([0.9, 0.1, 1.5])
    closest_point = get_closest_point(line_begin, line_end)
    line_length = np.linalg.norm(line_end - line_begin)
    line_direction = (line_end - line_begin) / line_length
    line = prepare_solid_angle_line_sampling(line_begin, line_direction, line_length)
    radius = 0.4
    min_distance = 1.0 / line.rcp_min_distance
    opening_angle = np.arcsin(radius / min_distance)
    line_normal = np.cross(line_direction, closest_point)
    # Set up a perspective projection
    yaw_angle, pitch_angle = -0.15 * np.pi, 0.03 * np.pi
    yaw_angle += np.pi / 180.0
    perspective_projection = get_euler_projection(yaw_angle, pitch_angle)
    # Set up projections specific to the considered line
    st_rotation = np.vstack([line_direction, closest_point, line_normal])
    ut_rotation = np.vstack([line_normal, closest_point, line_direction])
    # Do the plotting
    projection_list = [
        ("perspective", perspective_projection),
        ("st", OrthogonalProjection(st_rotation)),
        ("ut", OrthogonalProjection(ut_rotation)),
    ]
    for name, projection in projection_list:
        plot_line_solid_angle_single(name, line, line_begin, line_end, radius,
                                     line_direction, line_normal, opening_angle, projection)


def plot_inverse_cdf():
    """
    Creates plots of the inverse CDF for various parameter values.
    """
    figure = pyplot.figure(figsize=get_figure_size(column_width, 1.5 * column_width))
    pyplot.subplot(aspect=1.0)
    gammas = np.asarray([-np.arccos(0.3) / np.pi, -0.25, 0.0, 0.25, 0.5]) * np.pi
    colors = ["C1", "C0", "C2", "C4", "C3"]
    for gamma, color in zip(gammas, colors):
        normal = np.asarray([np.cos(gamma), np.sin(gamma)])
        s = np.linspace(-normal[1], 1.0, 10001)
        t = np.sqrt(np.maximum(0.0, 1.0 - s * s))
        F = normal[0] * s * s + normal[1] * (t * s + np.arcsin(s))
        normal_strings = [re.sub(r"(\..)0$", r"\1", ("%.2f" % value).replace(".00", "")) for value in normal]
        pyplot.plot(F, s, label=r"$F^{-1}_{%s,%s}$" % tuple(normal_strings), color=color)
    pyplot.yticks(np.linspace(-1.0, 1.0, 5))
    pyplot.xlabel(r"Integral $F_{n_s,n_t}$")
    pyplot.ylabel(r"Sample coordinate $\omega_s$")
    pyplot.ylim(-1.04, 1.04)
    pyplot.legend(loc="lower right", frameon=False)
    pyplot.xticks([-0.5 * np.pi, -0.25 * np.pi, 0.0, 0.25 * np.pi, 0.5 * np.pi],
                  [r"$-\frac{\pi}{2}$", r"$-\frac{\pi}{4}$", r"$0$", r"$\frac{\pi}{4}$", r"$\frac{\pi}{2}$"])
    figure.savefig("../Graphics/line_cdf/all_plots.pdf", bbox_inches="tight", pad_inches=0.01)


def magnify_error_cornell():
    """
    Creates magnified insets for the Cornell box with error visualization and
    creates a color bar.
    """
    for file_path, name in list_experiments("cornell_box_error_.*.png"):
        screenshot = Screenshot(file_path)
        magnifier_list = [
            (1, (500, 306), "C3"),
            (2, (544, 330), "C1"),
            (0, (500, 170), "C5"),
        ]
        padding = 32
        zoom = 8
        size = (24, 24)
        for position, left_top, color in magnifier_list:
            inset_left_top = (1024 - padding - size[0] * zoom, padding + (size[1] * zoom + padding) * position)
            screenshot.add_magnifier(left_top, size, inset_left_top, zoom, color, 12)
        screenshot.save(path.join("../Graphics/line_error_cornell", name))
    # Plot the color map
    figure = pyplot.figure(figsize=get_figure_size(0.024 * column_width, 0.4 * column_width))
    axes = pyplot.subplot()
    pyplot.colorbar(cm.ScalarMappable(norm=colors.Normalize(-7.0, -2.0), cmap="tab20b"), orientation="vertical", format="$10^{%.0f}$", cax=axes)
    figure.savefig("../Graphics/line_error_cornell/color_map.pdf", bbox_inches="tight", pad_inches=0.02)


def plot_ltc_space(name, line_begin, line_end, world_to_cosine_space):
    """
    Implements plot_ltc() using the given transform to construct the linearly
    transformed cosine and to transform the line
    """
    closest_point = get_closest_point(line_begin, line_end)
    # Set up a perspective projection
    yaw_angle, pitch_angle = -0.15 * np.pi, 0.03 * np.pi
    yaw_angle += np.pi / 180.0
    perspective_rotation = np.asarray([
        [1.0,0.0,0.0],
        [0.0, np.cos(pitch_angle), -np.sin(pitch_angle)],
        [0.0, np.sin(pitch_angle), np.cos(pitch_angle)]])
    perspective_rotation = perspective_rotation @ np.asarray([
        [np.cos(yaw_angle), -np.sin(yaw_angle), 0.0],
        [0.0, 0.0, 1.0],
        [-np.sin(yaw_angle), -np.cos(yaw_angle), 0.0]])
    projection = OrthogonalProjection(perspective_rotation)
    # Plot a hemisphere
    figure = pyplot.figure(figsize=get_figure_size(column_width, column_width * 0.364))
    pyplot.subplot(aspect=1.0, xticks=[], yticks=[], frameon=False)
    hemisphere_color = np.asarray([80, 78, 134]) / 255.0
    plot_hemisphere(projection, 0, 1, color=hemisphere_color)
    half_phase = np.linspace(-0.5 * np.pi, 0.5 * np.pi, 512)
    pyplot.plot(np.sin(half_phase), np.cos(half_phase), color=hemisphere_color)
    # Plot the horizon
    normal_cosine_space = np.asarray([0.0, 0.0, 1.0])
    normal_world_space = world_to_cosine_space.T @ normal_cosine_space
    horizon_tangent = np.cross(projection.matrix[2, :], normal_world_space)
    horizon_tangent /= np.linalg.norm(horizon_tangent)
    horizon_bitangent = np.cross(horizon_tangent, normal_world_space)
    horizon_bitangent /= np.linalg.norm(horizon_bitangent)
    for sign, style in [(-1.0, "-"), (1.0, "--")]:
        phase = sign * np.linspace(0.0 * np.pi, 1.0 * np.pi, 512)[:, np.newaxis]
        horizon = np.cos(phase) * horizon_tangent[np.newaxis, :] + np.sin(phase) * horizon_bitangent[np.newaxis, :]
        projection.plot_curve(horizon, color="C4", linestyle=style)
    # Figure out the geometry of the transformed cylinder
    radius = 0.1
    line_normal = np.cross(line_begin, line_end)
    line_normal /= np.linalg.norm(line_normal)
    radius *= np.linalg.norm(np.linalg.inv(world_to_cosine_space).T @ line_normal) / np.linalg.norm(line_normal)
    print(radius)
    line_begin = world_to_cosine_space @ line_begin
    line_end = world_to_cosine_space @ line_end
    line_direction = (line_end - line_begin) / np.linalg.norm(line_end - line_begin)
    line = prepare_solid_angle_line_sampling(line_begin, line_direction, np.linalg.norm(line_end - line_begin))
    min_distance = 1.0 / line.rcp_min_distance
    opening_angle = np.arcsin(radius / min_distance)
    # Plot its solid angle (outline only)
    side_list = list()
    for sign in [-1, 1]:
        s = np.linspace(line.begin_s, line.begin_s + line.extent_s, 513)
        tu = np.sqrt(1.0 - s * s)
        t = np.cos(opening_angle) * tu
        u = sign * np.sin(opening_angle) * tu
        side = (s[:, np.newaxis] * line_direction[np.newaxis, :]
                + t[:, np.newaxis] * closest_point[np.newaxis, :]
                + u[:, np.newaxis] * line_normal[np.newaxis, :])
        side_list.append(side)
    boundary = np.vstack([side_list[0], side_list[1][::-1], side_list[0][0:1, :]])
    projection.plot_curve(boundary, color=0.9 * solid_angle_line_color, solid_capstyle="round", linewidth=1.75)
    projection.plot_curve(boundary, color=1.0 * solid_angle_line_color, solid_capstyle="round", linewidth=0.75)
    if name=="world_space":
        pyplot.annotate(r"$\mathbb{L}$", (-0.25, 0.85), color=3 * [1.0])
    else:
        pyplot.annotate(r"$M^{-1}(\omega_o)\mathbb{L}$", (-0.3, 0.65), color="k")
    # Shade the solid angle using rasterization
    x_min, x_max = (-1.0, 1.0)
    y_min, y_max = (-0.53, 1.0)
    dots_per_unit = 400.0
    width, height = (int((x_max - x_min) * dots_per_unit), int((y_max - y_min) * dots_per_unit))
    x, y = np.meshgrid(np.linspace(x_min, x_max, width), np.linspace(y_min, y_max, height))
    z = np.sqrt(1.0 - x * x - y * y)
    view_space = np.zeros(list(x.shape) + [3])
    view_space[:, :, 0] = x
    view_space[:, :, 1] = y
    view_space[:, :, 2] = z
    world_space = np.einsum("ij, ...j-> ...i", projection.matrix.T, view_space)
    cosine_space = np.einsum("ij, ...j-> ...i", world_to_cosine_space, world_space)
    cosine_space_length = np.linalg.norm(cosine_space, axis=-1)
    ltc_density = np.maximum(0.0, cosine_space[..., 2]) * np.linalg.det(world_to_cosine_space) / cosine_space_length**4
    ltc_density_rgb = cm.get_cmap("plasma")(ltc_density / np.nanmax(ltc_density))[...,0:3]
    ltc_density_rgb = np.where(np.isnan(ltc_density)[:, :, np.newaxis].repeat(3, 2), 1.0, ltc_density_rgb)
    ltc_density_rgb = np.where((cosine_space[..., 2] < 0.0)[:, :, np.newaxis].repeat(3, 2), 1.0, ltc_density_rgb)
    pyplot.imshow(ltc_density_rgb, extent=(x_min, x_max, y_min, y_max), zorder=-1000.0, origin="lower")
    padding = 0.01
    pyplot.xlim(-1.0 - padding, 1.0 + padding)
    pyplot.ylim(-0.53 - padding, 1.0 + padding)
    # Add a few annotations
    pyplot.annotate(r"$\int_\mathbb{L}$  " if name=="world_space" else r"$\int_{M^{-1}(\omega_o)\mathbb{L}}$", (-0.61, 0.9), horizontalalignment="right", verticalalignment="center")
    pyplot.annotate(r"d$\omega\;\,=$" if name=="world_space" else r"d$\omega_c$", (0.61, 0.9), horizontalalignment="left", verticalalignment="center")
    if name == "world_space":
        pyplot.annotate(r"$M(\omega_o)\mathbb{H}$", (0.2, -0.45), color=3 * [1.0])
    else:
        pyplot.annotate(r"$\mathbb{H}$", (0.4, -0.05), color=3 * [1.0])
    # Save as *.svg
    output_path = "../Graphics/line_ltc/%s" % name
    figure.savefig(output_path + ".svg", bbox_inches="tight", pad_inches=0.0)
    # Export the image as *.png without resampling
    imwrite(output_path + ".png", ltc_density_rgb)
    # Now patch the *.svg file using an image without resampling
    with open(output_path + ".svg", "r") as svg_file:
        svg_code = svg_file.read()
    svg_code = re.sub(r'data:image/png;base64,[^\"]+', output_path + ".png", svg_code, count=1)
    with open(output_path + ".svg", "w") as svg_file:
        svg_file.write(svg_code)
    # Convert from SVG to PDF using Inkscape
    # inkscape_path = "C:/Program files/Inkscape/inkscape.exe"
    inkscape_path = "inkscape"
    subprocess.call([inkscape_path, "--file", output_path + ".svg",
                     "--export-pdf", output_path + ".pdf"])


def plot_ltc():
    """
    Creates a plot showing how linearly transformed cosines (LTCs) are
    constructed and how they apply to linear lights.
    """
    shading_pitch = -0.8 * np.pi
    world_to_shading_space = np.asarray([
        [np.cos(shading_pitch), -np.sin(shading_pitch), 0.0],
        [np.sin(shading_pitch), np.cos(shading_pitch), 0.0],
        [0.0, 0.0, 1.0],])
    # The code to get this transform out of ltc_table.c is:
    # if (i==34 && j == (20*resolution+26))
    # printf("[[%f, %f, %f], [%f, %f, %f], [%f, %f, %f]]\n",
    #     inverse[0][0], inverse[0][1], inverse[0][2],
    #     inverse[1][0], inverse[1][1], inverse[1][2],
    #     inverse[2][0], inverse[2][1], inverse[2][2]);
    shading_to_cosine_space = np.asarray([[0.782347, 0.000000, -0.134090], [0.000000, 1.000000, 0.000000], [0.422520, 0.000000, 0.262691]]).T
    world_to_cosine_space = shading_to_cosine_space @ world_to_shading_space
    cosine_to_world_space = np.linalg.inv(world_to_cosine_space)
    line_begin = cosine_to_world_space @ np.asarray([-1.3, -0.3, 1.8])
    line_end = cosine_to_world_space @ np.asarray([0.9, 0.1, 1.5])
    plot_ltc_space("cosine_space", line_begin, line_end, np.eye(3))
    plot_ltc_space("world_space", line_begin, line_end, world_to_cosine_space)


def magnify_diffuse_cornell():
    """
    Creates magnified insets for the Cornell box with diffuse shading.
    """
    for file_path, name in list_experiments("cornell_box_.*.png"):
        screenshot = Screenshot(file_path)
        magnifier_list = [
            (1, (363, 336), "C0"),
            (2, (401, 488), "C1"),
            (0, (143, 355), "C2"),
            (3, (614, 966), "C3"),
        ]
        padding = 32
        zoom = 8
        size = (24, 24)
        for position, left_top, color in magnifier_list:
            inset_left_top = (1024 - padding - size[0] * zoom, padding + (size[1] * zoom + padding) * position)
            screenshot.add_magnifier(left_top, size, inset_left_top, zoom, color, 15)
        screenshot.add_magnifier((308, 180), size, (padding, padding), zoom, "C4", 15)
        screenshot.save(path.join("../Graphics/line_diffuse_cornell", name))


def magnify_stratification_comparison():
    """
    Creates magnified insets for the Bistro screenshots that compare white and
    blue noise and independent sampling versus uniform jittered sampling
    """
    for file_path, name in list_experiments("bistro_inside_.*.png"):
        screenshot = Screenshot(file_path)
        padding = 32
        zoom = 8
        size = (48, 48)
        small_size = (32, 32)
        border = 16
        screenshot.add_magnifier((822, 43), size, (padding, padding + 2 * border), zoom, "C2", border)
        screenshot.add_magnifier((814, 708), size, (padding, screenshot.image.shape[0] - padding - size[1] * zoom),
                                 zoom, "C1", border)
        screenshot.add_magnifier((534, 786), small_size, (1280 - padding - small_size[0] * zoom, padding + 2 * border),
                                 zoom, "C3", border)
        screenshot.save(path.join("../Graphics/line_stratification_comparison", name))


def print_timing_table(scene_name):
    """
    Prints a table with timings for line sampling techniques tested on the scene
    with the given name.
    """
    print("Printing timings for %s:" % scene_name)
    # Read in the timings
    timing_dict = dict()
    for file_name in listdir("../vulkan_renderer/data/experiments"):
        match = re.match(r"timings_%s_([0-9]+)_(\w+)_([0-9]+\.[0-9]+)\.png" % scene_name, file_name)
        if match is not None:
            timing_dict[int(match.group(1)), match.group(2)] = float(match.group(3))
    # The order in which techniques are listed
    techniques = [
        "area",
        "solid_angle",
        "clipped_solid_angle",
        "linear_cosine_warp_hart",
        "quadratic_cosine_warp_hart",
        "projected_solid_angle_li",
        "projected_solid_angle",
    ]
    def format_timing(timing_in_milliseconds):
        timing = timing_in_milliseconds
        if timing < 1.0:
            return "%.2f" % timing
        elif timing < 10.0:
            return "%.2f" % timing
        elif timing < 100.0:
            return "%.1f" % timing
        else:
            return "%.0f" % timing

    # Print them line by line
    print_name = False
    for technique in techniques:
        if print_name:
            print(technique, end="\t")
        for light_count in [128, 1]:
            print("%s\t" % format_timing(timing_dict[light_count, technique] - timing_dict[light_count, "baseline"]), end="")
        print()
    if print_name:
        print("baseline", end="\t")
    for light_count in [128, 1]:
        print("%s\t" % format_timing(timing_dict[light_count, "baseline"]), end="")
    print()


if __name__ == "__main__":
    # magnify_teaser()
    # plot_line_solid_angle()
    # plot_inverse_cdf()
    # magnify_error_cornell()
    # plot_ltc()
    # magnify_diffuse_cornell()
    # magnify_stratification_comparison()
    # print_timing_table("roughness_planes")
    # print_timing_table("Bistro_outside")
    # print_experiment_rmses("bistro_outside_.*.hdr", "bistro_outside_reference.hdr")
    # print_experiment_rmses("cornell_box_.*.hdr", "cornell_box_reference.hdr")
    # print_experiment_rmses("bistro_inside_.*.hdr", "bistro_inside_reference.hdr")
    # pyplot.show()
    # print_run_time_tables()
    print("Done.")
