#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

def iz0(lhs, rhs):
    """
    :return: The intersection of the line connecting the given two 3D points
    with the plane z == 0.0.
    """
    lerp = lhs[2] / (lhs[2] - rhs[2])
    return (1.0 - lerp) * lhs + lerp * rhs


def clip_polygon(polygon, max_vertex_count):
    """
    This function clips a convex polygon to the upper hemisphere (i.e. the
    half-space with non-negative z-coordinate).
    :param polygon: Array of shape (vertex_count, 3) providing vertex locations
    :param max_vertex_count: Size of the output array. At least
        vertex_count + 1.
    :return: A pair (new_vertex_count, clipped_polygon) where clipped_polygon
        is an array of shape (max_vertex_count, 3) providing the clipped
        polygon. If new_vertex_count < max_vertex_count, the first row is
        repeated at the end.
    """
    polygon = np.asarray(polygon)
    # Encode the whole configuration into a single integer
    vertex_count = polygon.shape[0]
    bit_mask = vertex_count
    for i in range(vertex_count):
        bit_mask |= (1 << (i + 3)) if polygon[i, 2] > 0.0 else 0
    # Make space for extra vertices
    v = np.zeros((max_vertex_count, 3))
    v[0:vertex_count, :] = polygon
    # This code is autogenerated. It is fairly pointless in Python but
    # justified from a performance standpoint in GLSL and this version is just
    # a port of that for debugging and analysis.
    # AUTOGENERATED PART BEGIN
    if max_vertex_count >= 4:
        if bit_mask ==    3:   vc = 0;
        if bit_mask ==   59:   vc = 3;   v[3] = v[0]
        if bit_mask ==   11:   vc = 3;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[2, :], v[0, :]);   v[3, :] = v[0, :]
        if bit_mask ==   19:   vc = 3;   v[0, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
        if bit_mask ==   35:   vc = 3;   v[0, :] = iz0(v[2, :], v[0, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
    if max_vertex_count == 4:
        if bit_mask ==   27:   vc = 4;   v[3, :] = iz0(v[2, :], v[0, :]);   v[2, :] = iz0(v[1, :], v[2, :])
        if bit_mask ==   51:   vc = 4;   v[3, :] = iz0(v[2, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :])
        if bit_mask ==   43:   vc = 4;   v[3, :] = v[2, :];   v[2, :] = iz0(v[1, :], v[2, :]);   v[1, :] = iz0(v[0, :], v[1, :])
    elif max_vertex_count > 4:
        if bit_mask ==   27:   vc = 4;   v[3, :] = iz0(v[2, :], v[0, :]);   v[2, :] = iz0(v[1, :], v[2, :]);   v[4, :] = v[0, :]
        if bit_mask ==   51:   vc = 4;   v[3, :] = iz0(v[2, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :]);   v[4, :] = v[0, :]
        if bit_mask ==   43:   vc = 4;   v[3, :] = v[2, :];   v[2, :] = iz0(v[1, :], v[2, :]);   v[1, :] = iz0(v[0, :], v[1, :]);   v[4, :] = v[0, :]
    if max_vertex_count >= 5:
        if bit_mask ==    4:   vc = 0;
        if bit_mask ==  124:   vc = 4;   v[4] = v[0]
        if bit_mask ==   12:   vc = 3;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[3, :], v[0, :]);   v[3, :] = v[0, :]
        if bit_mask ==   20:   vc = 3;   v[0, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
        if bit_mask ==   36:   vc = 3;   v[0, :] = iz0(v[2, :], v[3, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
        if bit_mask ==   68:   vc = 3;   v[1, :] = iz0(v[3, :], v[0, :]);   v[0, :] = v[3, :];   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==   28:   vc = 4;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[3, :], v[0, :]);   v[4, :] = v[0, :]
        if bit_mask ==   52:   vc = 4;   v[0, :] = iz0(v[0, :], v[1, :]);   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = v[0, :]
        if bit_mask ==  100:   vc = 4;   v[0, :] = iz0(v[3, :], v[0, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[4, :] = v[0, :]
        if bit_mask ==   76:   vc = 4;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[2, :], v[3, :]);   v[4, :] = v[0, :]
    if max_vertex_count == 5:
        if bit_mask ==   60:   vc = 5;   v[4, :] = iz0(v[3, :], v[0, :]);   v[3, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  116:   vc = 5;   v[4, :] = iz0(v[3, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :])
        if bit_mask ==  108:   vc = 5;   v[4, :] = v[0, :];   v[0, :] = iz0(v[0, :], v[1, :]);   v[1, :] = iz0(v[1, :], v[2, :])
        if bit_mask ==   92:   vc = 5;   v[4, :] = v[3, :];   v[3, :] = iz0(v[2, :], v[3, :]);   v[2, :] = iz0(v[1, :], v[2, :])
    elif max_vertex_count > 5:
        if bit_mask ==   60:   vc = 5;   v[4, :] = iz0(v[3, :], v[0, :]);   v[3, :] = iz0(v[2, :], v[3, :]);   v[5, :] = v[0, :]
        if bit_mask ==  116:   vc = 5;   v[4, :] = iz0(v[3, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :]);   v[5, :] = v[0, :]
        if bit_mask ==  108:   vc = 5;   v[4, :] = v[0, :];   v[0, :] = iz0(v[0, :], v[1, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[5, :] = v[0, :]
        if bit_mask ==   92:   vc = 5;   v[4, :] = v[3, :];   v[3, :] = iz0(v[2, :], v[3, :]);   v[2, :] = iz0(v[1, :], v[2, :]);   v[5, :] = v[0, :]
    if max_vertex_count >= 6:
        if bit_mask ==    5:   vc = 0;
        if bit_mask ==  253:   vc = 5;   v[5] = v[0]
        if bit_mask ==   13:   vc = 3;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[4, :], v[0, :]);   v[3, :] = v[0, :]
        if bit_mask ==   21:   vc = 3;   v[0, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
        if bit_mask ==   37:   vc = 3;   v[0, :] = iz0(v[2, :], v[3, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
        if bit_mask ==   69:   vc = 3;   v[0, :] = v[3, :];   v[1, :] = iz0(v[3, :], v[4, :]);   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  133:   vc = 3;   v[1, :] = v[4, :];   v[2, :] = iz0(v[4, :], v[0, :]);   v[0, :] = iz0(v[3, :], v[4, :]);   v[3, :] = v[0, :]
        if bit_mask ==   29:   vc = 4;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[4, :], v[0, :]);   v[4, :] = v[0, :]
        if bit_mask ==   53:   vc = 4;   v[0, :] = iz0(v[0, :], v[1, :]);   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = v[0, :]
        if bit_mask ==  101:   vc = 4;   v[0, :] = iz0(v[3, :], v[4, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[4, :] = v[0, :]
        if bit_mask ==  197:   vc = 4;   v[1, :] = iz0(v[4, :], v[0, :]);   v[0, :] = v[4, :];   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  141:   vc = 4;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[3, :], v[4, :]);   v[3, :] = v[4, :];   v[4, :] = v[0, :]
        if bit_mask ==   61:   vc = 5;   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = iz0(v[4, :], v[0, :]);   v[5, :] = v[0, :]
        if bit_mask ==  117:   vc = 5;   v[0, :] = iz0(v[0, :], v[1, :]);   v[4, :] = iz0(v[3, :], v[4, :]);   v[5, :] = v[0, :]
        if bit_mask ==  229:   vc = 5;   v[0, :] = iz0(v[4, :], v[0, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[5, :] = v[0, :]
        if bit_mask ==  205:   vc = 5;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[2, :], v[3, :]);   v[5, :] = v[0, :]
        if bit_mask ==  157:   vc = 5;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[3, :], v[4, :]);   v[5, :] = v[0, :]
    if max_vertex_count == 6:
        if bit_mask ==  125:   vc = 6;   v[5, :] = iz0(v[4, :], v[0, :]);   v[4, :] = iz0(v[3, :], v[4, :])
        if bit_mask ==  245:   vc = 6;   v[5, :] = iz0(v[4, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :])
        if bit_mask ==  237:   vc = 6;   v[5, :] = v[0, :];   v[0, :] = iz0(v[0, :], v[1, :]);   v[1, :] = iz0(v[1, :], v[2, :])
        if bit_mask ==  221:   vc = 6;   v[5, :] = v[4, :];   v[4, :] = v[3, :];   v[3, :] = iz0(v[2, :], v[3, :]);   v[2, :] = iz0(v[1, :], v[2, :])
        if bit_mask ==  189:   vc = 6;   v[5, :] = v[4, :];   v[4, :] = iz0(v[3, :], v[4, :]);   v[3, :] = iz0(v[2, :], v[3, :])
    elif max_vertex_count > 6:
        if bit_mask ==  125:   vc = 6;   v[5, :] = iz0(v[4, :], v[0, :]);   v[4, :] = iz0(v[3, :], v[4, :]);   v[6, :] = v[0, :]
        if bit_mask ==  245:   vc = 6;   v[5, :] = iz0(v[4, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :]);   v[6, :] = v[0, :]
        if bit_mask ==  237:   vc = 6;   v[5, :] = v[0, :];   v[0, :] = iz0(v[0, :], v[1, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[6, :] = v[0, :]
        if bit_mask ==  221:   vc = 6;   v[5, :] = v[4, :];   v[4, :] = v[3, :];   v[3, :] = iz0(v[2, :], v[3, :]);   v[2, :] = iz0(v[1, :], v[2, :]);   v[6, :] = v[0, :]
        if bit_mask ==  189:   vc = 6;   v[5, :] = v[4, :];   v[4, :] = iz0(v[3, :], v[4, :]);   v[3, :] = iz0(v[2, :], v[3, :]);   v[6, :] = v[0, :]
    if max_vertex_count >= 7:
        if bit_mask ==    6:   vc = 0;
        if bit_mask ==  510:   vc = 6;   v[6] = v[0]
        if bit_mask ==   14:   vc = 3;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[5, :], v[0, :]);   v[3, :] = v[0, :]
        if bit_mask ==   22:   vc = 3;   v[0, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
        if bit_mask ==   38:   vc = 3;   v[0, :] = iz0(v[2, :], v[3, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
        if bit_mask ==   70:   vc = 3;   v[0, :] = v[3, :];   v[1, :] = iz0(v[3, :], v[4, :]);   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  134:   vc = 3;   v[0, :] = iz0(v[3, :], v[4, :]);   v[1, :] = v[4, :];   v[2, :] = iz0(v[4, :], v[5, :]);   v[3, :] = v[0, :]
        if bit_mask ==  262:   vc = 3;   v[1, :] = v[5, :];   v[2, :] = iz0(v[5, :], v[0, :]);   v[0, :] = iz0(v[4, :], v[5, :]);   v[3, :] = v[0, :]
        if bit_mask ==   30:   vc = 4;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[5, :], v[0, :]);   v[4, :] = v[0, :]
        if bit_mask ==   54:   vc = 4;   v[0, :] = iz0(v[0, :], v[1, :]);   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = v[0, :]
        if bit_mask ==  102:   vc = 4;   v[0, :] = iz0(v[3, :], v[4, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[4, :] = v[0, :]
        if bit_mask ==  198:   vc = 4;   v[0, :] = v[4, :];   v[1, :] = iz0(v[4, :], v[5, :]);   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  390:   vc = 4;   v[1, :] = v[5, :];   v[2, :] = iz0(v[5, :], v[0, :]);   v[0, :] = v[4, :];   v[3, :] = iz0(v[3, :], v[4, :])
        if bit_mask ==  270:   vc = 4;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[4, :], v[5, :]);   v[3, :] = v[5, :];   v[4, :] = v[0, :]
        if bit_mask ==   62:   vc = 5;   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = iz0(v[5, :], v[0, :]);   v[5, :] = v[0, :]
        if bit_mask ==  118:   vc = 5;   v[0, :] = iz0(v[0, :], v[1, :]);   v[4, :] = iz0(v[3, :], v[4, :]);   v[5, :] = v[0, :]
        if bit_mask ==  230:   vc = 5;   v[0, :] = iz0(v[4, :], v[5, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[5, :] = v[0, :]
        if bit_mask ==  454:   vc = 5;   v[1, :] = iz0(v[5, :], v[0, :]);   v[0, :] = v[5, :];   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  398:   vc = 5;   v[2, :] = iz0(v[0, :], v[1, :]);   v[1, :] = v[0, :];   v[0, :] = v[5, :];   v[3, :] = iz0(v[3, :], v[4, :])
        if bit_mask ==  286:   vc = 5;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[4, :], v[5, :]);   v[4, :] = v[5, :];   v[5, :] = v[0, :]
        if bit_mask ==  126:   vc = 6;   v[4, :] = iz0(v[3, :], v[4, :]);   v[5, :] = iz0(v[5, :], v[0, :]);   v[6, :] = v[0, :]
        if bit_mask ==  246:   vc = 6;   v[0, :] = iz0(v[0, :], v[1, :]);   v[5, :] = iz0(v[4, :], v[5, :]);   v[6, :] = v[0, :]
        if bit_mask ==  486:   vc = 6;   v[0, :] = iz0(v[5, :], v[0, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[6, :] = v[0, :]
        if bit_mask ==  462:   vc = 6;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[2, :], v[3, :]);   v[6, :] = v[0, :]
        if bit_mask ==  414:   vc = 6;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[3, :], v[4, :]);   v[6, :] = v[0, :]
        if bit_mask ==  318:   vc = 6;   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = iz0(v[4, :], v[5, :]);   v[6, :] = v[0, :]
    if max_vertex_count == 7:
        if bit_mask ==  254:   vc = 7;   v[6, :] = iz0(v[5, :], v[0, :]);   v[5, :] = iz0(v[4, :], v[5, :])
        if bit_mask ==  502:   vc = 7;   v[6, :] = iz0(v[5, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :])
        if bit_mask ==  494:   vc = 7;   v[6, :] = v[0, :];   v[0, :] = iz0(v[0, :], v[1, :]);   v[1, :] = iz0(v[1, :], v[2, :])
        if bit_mask ==  478:   vc = 7;   v[6, :] = v[0, :];   v[0, :] = v[1, :];   v[1, :] = iz0(v[1, :], v[2, :]);   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  446:   vc = 7;   v[6, :] = v[5, :];   v[5, :] = v[4, :];   v[4, :] = iz0(v[3, :], v[4, :]);   v[3, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  382:   vc = 7;   v[6, :] = v[5, :];   v[5, :] = iz0(v[4, :], v[5, :]);   v[4, :] = iz0(v[3, :], v[4, :])
    elif max_vertex_count > 7:
        if bit_mask ==  254:   vc = 7;   v[6, :] = iz0(v[5, :], v[0, :]);   v[5, :] = iz0(v[4, :], v[5, :]);   v[7, :] = v[0, :]
        if bit_mask ==  502:   vc = 7;   v[6, :] = iz0(v[5, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :]);   v[7, :] = v[0, :]
        if bit_mask ==  494:   vc = 7;   v[6, :] = v[0, :];   v[0, :] = iz0(v[0, :], v[1, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[7, :] = v[0, :]
        if bit_mask ==  478:   vc = 7;   v[6, :] = v[0, :];   v[0, :] = v[1, :];   v[1, :] = iz0(v[1, :], v[2, :]);   v[2, :] = iz0(v[2, :], v[3, :]);   v[7, :] = v[0, :]
        if bit_mask ==  446:   vc = 7;   v[6, :] = v[5, :];   v[5, :] = v[4, :];   v[4, :] = iz0(v[3, :], v[4, :]);   v[3, :] = iz0(v[2, :], v[3, :]);   v[7, :] = v[0, :]
        if bit_mask ==  382:   vc = 7;   v[6, :] = v[5, :];   v[5, :] = iz0(v[4, :], v[5, :]);   v[4, :] = iz0(v[3, :], v[4, :]);   v[7, :] = v[0, :]
    if max_vertex_count >= 8:
        if bit_mask ==    7:   vc = 0;
        if bit_mask == 1023:   vc = 7;   v[7] = v[0]
        if bit_mask ==   15:   vc = 3;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[6, :], v[0, :]);   v[3, :] = v[0, :]
        if bit_mask ==   23:   vc = 3;   v[0, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
        if bit_mask ==   39:   vc = 3;   v[0, :] = iz0(v[2, :], v[3, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[3, :] = v[0, :]
        if bit_mask ==   71:   vc = 3;   v[0, :] = v[3, :];   v[1, :] = iz0(v[3, :], v[4, :]);   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  135:   vc = 3;   v[0, :] = iz0(v[3, :], v[4, :]);   v[1, :] = v[4, :];   v[2, :] = iz0(v[4, :], v[5, :]);   v[3, :] = v[0, :]
        if bit_mask ==  263:   vc = 3;   v[0, :] = iz0(v[4, :], v[5, :]);   v[1, :] = v[5, :];   v[2, :] = iz0(v[5, :], v[6, :]);   v[3, :] = v[0, :]
        if bit_mask ==  519:   vc = 3;   v[1, :] = v[6, :];   v[2, :] = iz0(v[6, :], v[0, :]);   v[0, :] = iz0(v[5, :], v[6, :]);   v[3, :] = v[0, :]
        if bit_mask ==   31:   vc = 4;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[6, :], v[0, :]);   v[4, :] = v[0, :]
        if bit_mask ==   55:   vc = 4;   v[0, :] = iz0(v[0, :], v[1, :]);   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = v[0, :]
        if bit_mask ==  103:   vc = 4;   v[0, :] = iz0(v[3, :], v[4, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[4, :] = v[0, :]
        if bit_mask ==  199:   vc = 4;   v[0, :] = v[4, :];   v[1, :] = iz0(v[4, :], v[5, :]);   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  391:   vc = 4;   v[0, :] = v[4, :];   v[1, :] = v[5, :];   v[2, :] = iz0(v[5, :], v[6, :]);   v[3, :] = iz0(v[3, :], v[4, :])
        if bit_mask ==  775:   vc = 4;   v[1, :] = v[5, :];   v[2, :] = v[6, :];   v[3, :] = iz0(v[6, :], v[0, :]);   v[0, :] = iz0(v[4, :], v[5, :]);   v[4, :] = v[0, :]
        if bit_mask ==  527:   vc = 4;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[5, :], v[6, :]);   v[3, :] = v[6, :];   v[4, :] = v[0, :]
        if bit_mask ==   63:   vc = 5;   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = iz0(v[6, :], v[0, :]);   v[5, :] = v[0, :]
        if bit_mask ==  119:   vc = 5;   v[0, :] = iz0(v[0, :], v[1, :]);   v[4, :] = iz0(v[3, :], v[4, :]);   v[5, :] = v[0, :]
        if bit_mask ==  231:   vc = 5;   v[0, :] = iz0(v[4, :], v[5, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[5, :] = v[0, :]
        if bit_mask ==  455:   vc = 5;   v[0, :] = v[5, :];   v[1, :] = iz0(v[5, :], v[6, :]);   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  903:   vc = 5;   v[1, :] = v[6, :];   v[2, :] = iz0(v[6, :], v[0, :]);   v[0, :] = v[5, :];   v[3, :] = iz0(v[3, :], v[4, :])
        if bit_mask ==  783:   vc = 5;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[4, :], v[5, :]);   v[3, :] = v[5, :];   v[4, :] = v[6, :];   v[5, :] = v[0, :]
        if bit_mask ==  543:   vc = 5;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[5, :], v[6, :]);   v[4, :] = v[6, :];   v[5, :] = v[0, :]
        if bit_mask ==  127:   vc = 6;   v[4, :] = iz0(v[3, :], v[4, :]);   v[5, :] = iz0(v[6, :], v[0, :]);   v[6, :] = v[0, :]
        if bit_mask ==  247:   vc = 6;   v[0, :] = iz0(v[0, :], v[1, :]);   v[5, :] = iz0(v[4, :], v[5, :]);   v[6, :] = v[0, :]
        if bit_mask ==  487:   vc = 6;   v[0, :] = iz0(v[5, :], v[6, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[6, :] = v[0, :]
        if bit_mask ==  967:   vc = 6;   v[1, :] = iz0(v[6, :], v[0, :]);   v[0, :] = v[6, :];   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  911:   vc = 6;   v[2, :] = iz0(v[0, :], v[1, :]);   v[1, :] = v[0, :];   v[0, :] = v[6, :];   v[3, :] = iz0(v[3, :], v[4, :])
        if bit_mask ==  799:   vc = 6;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[4, :], v[5, :]);   v[4, :] = v[5, :];   v[5, :] = v[6, :];   v[6, :] = v[0, :]
        if bit_mask ==  575:   vc = 6;   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = iz0(v[5, :], v[6, :]);   v[5, :] = v[6, :];   v[6, :] = v[0, :]
        if bit_mask ==  255:   vc = 7;   v[5, :] = iz0(v[4, :], v[5, :]);   v[6, :] = iz0(v[6, :], v[0, :]);   v[7, :] = v[0, :]
        if bit_mask ==  503:   vc = 7;   v[0, :] = iz0(v[0, :], v[1, :]);   v[6, :] = iz0(v[5, :], v[6, :]);   v[7, :] = v[0, :]
        if bit_mask ==  999:   vc = 7;   v[0, :] = iz0(v[6, :], v[0, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[7, :] = v[0, :]
        if bit_mask ==  975:   vc = 7;   v[1, :] = iz0(v[0, :], v[1, :]);   v[2, :] = iz0(v[2, :], v[3, :]);   v[7, :] = v[0, :]
        if bit_mask ==  927:   vc = 7;   v[2, :] = iz0(v[1, :], v[2, :]);   v[3, :] = iz0(v[3, :], v[4, :]);   v[7, :] = v[0, :]
        if bit_mask ==  831:   vc = 7;   v[3, :] = iz0(v[2, :], v[3, :]);   v[4, :] = iz0(v[4, :], v[5, :]);   v[7, :] = v[0, :]
        if bit_mask ==  639:   vc = 7;   v[4, :] = iz0(v[3, :], v[4, :]);   v[5, :] = iz0(v[5, :], v[6, :]);   v[7, :] = v[0, :]
    if max_vertex_count == 8:
        if bit_mask ==  511:   vc = 8;   v[7, :] = iz0(v[6, :], v[0, :]);   v[6, :] = iz0(v[5, :], v[6, :])
        if bit_mask == 1015:   vc = 8;   v[7, :] = iz0(v[6, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :])
        if bit_mask == 1007:   vc = 8;   v[7, :] = v[0, :];   v[0, :] = iz0(v[0, :], v[1, :]);   v[1, :] = iz0(v[1, :], v[2, :])
        if bit_mask ==  991:   vc = 8;   v[7, :] = v[0, :];   v[0, :] = v[1, :];   v[1, :] = iz0(v[1, :], v[2, :]);   v[2, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  959:   vc = 8;   v[7, :] = v[6, :];   v[6, :] = v[5, :];   v[5, :] = v[4, :];   v[4, :] = iz0(v[3, :], v[4, :]);   v[3, :] = iz0(v[2, :], v[3, :])
        if bit_mask ==  895:   vc = 8;   v[7, :] = v[6, :];   v[6, :] = v[5, :];   v[5, :] = iz0(v[4, :], v[5, :]);   v[4, :] = iz0(v[3, :], v[4, :])
        if bit_mask ==  767:   vc = 8;   v[7, :] = v[6, :];   v[6, :] = iz0(v[5, :], v[6, :]);   v[5, :] = iz0(v[4, :], v[5, :])
    elif max_vertex_count > 8:
        if bit_mask ==  511:   vc = 8;   v[7, :] = iz0(v[6, :], v[0, :]);   v[6, :] = iz0(v[5, :], v[6, :]);   v[8, :] = v[0, :]
        if bit_mask == 1015:   vc = 8;   v[7, :] = iz0(v[6, :], v[0, :]);   v[0, :] = iz0(v[0, :], v[1, :]);   v[8, :] = v[0, :]
        if bit_mask == 1007:   vc = 8;   v[7, :] = v[0, :];   v[0, :] = iz0(v[0, :], v[1, :]);   v[1, :] = iz0(v[1, :], v[2, :]);   v[8, :] = v[0, :]
        if bit_mask ==  991:   vc = 8;   v[7, :] = v[0, :];   v[0, :] = v[1, :];   v[1, :] = iz0(v[1, :], v[2, :]);   v[2, :] = iz0(v[2, :], v[3, :]);   v[8, :] = v[0, :]
        if bit_mask ==  959:   vc = 8;   v[7, :] = v[6, :];   v[6, :] = v[5, :];   v[5, :] = v[4, :];   v[4, :] = iz0(v[3, :], v[4, :]);   v[3, :] = iz0(v[2, :], v[3, :]);   v[8, :] = v[0, :]
        if bit_mask ==  895:   vc = 8;   v[7, :] = v[6, :];   v[6, :] = v[5, :];   v[5, :] = iz0(v[4, :], v[5, :]);   v[4, :] = iz0(v[3, :], v[4, :]);   v[8, :] = v[0, :]
        if bit_mask ==  767:   vc = 8;   v[7, :] = v[6, :];   v[6, :] = iz0(v[5, :], v[6, :]);   v[5, :] = iz0(v[4, :], v[5, :]);   v[8, :] = v[0, :]
 # AUTOGENERATED PART END
    return vc, v
