#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


# -*- coding: utf-8 -*-
import re
from slide_utility import *
from subprocess import Popen
from os import path, listdir
from line_sampling_ported import prepare_solid_angle_line_sampling, sample_solid_angle_line, \
    prepare_projected_solid_angle_line_sampling_without_clipping, sample_projected_solid_angle_line
from generate_graphics import solid_angle_line_color, light_color, get_closest_point, solid_angle_color, plot_axis
from hemisphere_plot_utility import get_euler_projection, OrthogonalProjection, plot_hemisphere, plot_annotated_point, \
    plot_annotated_length, plot_right_triangle
from polygon_clipping import clip_polygon
from matplotlib import cm, rcParams
from imageio import imread
from scipy import spatial


horizon_color = np.asarray(3 * [0.85 * 0.9])


def get_random_numbers(sample_count=16):
    """
    :return: An array of shape (sample_count,) holding random numbers between
        zero and one. Currently slightly jittered equidistant samples.
    """
    return np.linspace(0.0, 1.0 - 1.0 / sample_count, sample_count) + (0.35 + 0.3 * np.random.rand(sample_count)) / sample_count


def title_slide(_step):
    pyplot.text(8.0, 5.5, "BRDF Importance Sampling for Linear Lights", ha="center", fontsize=23.0)
    pyplot.text(8.0, 4.2, "Christoph Peters", ha="center")
    pyplot.text(8.0, 3.5, "Karlsruhe Institute of Technology", ha="center")
    pyplot.text(8.0, 2.8, "High-Performance Graphics 2021 (CGF)", ha="center")
    pyplot.text(8.0, 2.1, "2021-07-07", ha="center")


class GameSlide(Slide):
    """Shows a game screenshot to motivate the problem."""

    def __init__(self):
        super().__init__()
        self.image = imread("../Graphics/line_talk/resident_evil_7_linear_light_full.webp")

    def draw(self):
        draw_image(self.image, (0.0, 0.0), 16.0)
        pyplot.text(0.6, 8.1, "Resident Evil 7  (© Capcom)", fontsize=18.0, c="w", zorder=10000.0)


def sample_edges(vertices, close_loop):
    """
    This method is useful for plotting the solid angle of a polygon or other
    arcs of great circles.
    :param vertices: Array of shape (vertex_count, 3) providing vertices of the
        edges. Any two subsequent vertices get connected by an edge. Need not be
        normalized.
    :param close_loop: Pass True to ensure that the first and last vertex are
        connected by an edge.
    :return: An array of shape (sample_count, 3) providing samples of all
        specified edges as points on the unit sphere.
    """
    vertices = np.asarray(vertices)
    lerp = np.linspace(0.0, 1.0, 513)[:, np.newaxis]
    if close_loop:
        vertices = np.vstack([vertices, vertices[0:1, :]])
    vertices /= np.linalg.norm(vertices, axis=1)[:, np.newaxis]
    edges = [(1.0 - lerp) * vertices[j, np.newaxis, :] + lerp * vertices[j + 1, np.newaxis, :] for j in range(vertices.shape[0] - 1)]
    result = np.concatenate(edges)
    result /= np.linalg.norm(result, axis=-1)[:, np.newaxis]
    return result


def plot_line_solid_angle_talk(elements, line_begin, line_end, radius, projection, axes_origin=(-2.4, 0.0, 0.0)):
    """
    Visualizes a cylinder light in relation to a shading point and how its solid
    angle can be computed.
    :param elements: An iterable of strings containing any subset of:
    - hemisphere: A hemisphere around the shading point
    - solid_angle: The solid angle of the cylinder
    - endpoints: Labels for center points of the caps
    - radius: Label for the radius of the cylinder
    - closest_point: Label for the closest point on the line
    - triangle_end: A right triangle between origin, closest point and line end
    - triangle_boundary: A right triangle between origin, closest point and with
        a tangent to the cylinder as edge (all in the ut-plane)
    - extent_s: The extent of the cylinder solid angle along the s-axis
    - extent_t: The extent of the cylinder along the t-axis (diameter)
    - axes_stu: The used local coordinate frame. Omit letters to omit axes.
    :param line_begin: Center of the beginning cap of the cylinder (shape (3,))
    :param line_end: Center of the ending cap of the cylinder (shape (3,))
    :param radius: The radius of the cylinder.
    :param projection: The used OrthogonalProjection. If extent_s or extent_t
        are shown it must be a properly axis-aligned projection.
    :param axes_origin: The origin of the shown coordinate frame.
    """
    previous_size = rcParams["font.size"]
    rcParams["font.size"] = 9

    line_length = np.linalg.norm(line_end - line_begin)
    line_direction = (line_end - line_begin) / np.linalg.norm(line_end - line_begin)
    closest_point = get_closest_point(line_begin, line_end)
    line_normal = np.cross(line_direction, closest_point)
    line = prepare_solid_angle_line_sampling(line_begin, line_direction, line_length)
    min_distance = 1.0 / line.rcp_min_distance
    opening_angle = np.arcsin(radius / min_distance)

    # Plot a hemisphere
    def in_solid_angle(position):
        s = np.einsum("i, ...i -> ...", line_direction, position)
        t = np.einsum("i, ...i -> ...", closest_point, position)
        u = np.einsum("i, ...i -> ...", line_normal, position)
        angle = np.arctan2(u, t)
        return (s > line.begin_s) & (s < line.begin_s + line.extent_s) & (np.abs(angle) < opening_angle)

    def covered_by_solid_angle(position):
        depth = projection.transform(position)[..., 2]
        row = projection.matrix[2, :]
        reverse_position = position - 2.0 * depth[..., np.newaxis] * row.reshape([1] * (position.ndim - 1) + [3])
        return in_solid_angle(position) | in_solid_angle(reverse_position)
    if "hemisphere" in elements:
        if "solid_angle" in elements:
            plot_hemisphere(projection, 8, 4, covered_by_solid_angle, color=solid_angle_line_color)
            plot_hemisphere(projection, 8, 4, lambda position: ~covered_by_solid_angle(position))
        else:
            plot_hemisphere(projection, 8, 4)
        # Mark the shading point
        pyplot.scatter([0.0], [0.0], color="k", s=4.0, zorder=-35.0)
    # Plot the cylinder
    boundary_offset_direction = np.cross(line_direction, projection.matrix[2, :])
    boundary_offset_direction /= np.linalg.norm(boundary_offset_direction)
    boundary_offset_normal = np.cross(boundary_offset_direction, line_direction)
    if "cylinder" in elements:
        phase = np.linspace(-np.pi, np.pi, 513)
        bottom_line = list()
        top_line = list()
        cap_list = list()
        for normal_sign, cap_center in [(-1.0, line_begin), (1.0, line_end)]:
            cap = (cap_center[np.newaxis, :]
                   + line_normal * np.cos(phase)[:, np.newaxis] * radius
                   + closest_point[np.newaxis, :] * np.sin(phase)[:, np.newaxis] * radius)
            cap_list.append(cap)
            # It may be necessary to plot only half the cap
            if np.dot(line_direction * normal_sign, projection.matrix[2, :]) > 0.0:
                projection.plot_curve(cap, color="k")
            else:
                halves = [(-1.0, 0.85 * light_color, "--", -300.0), (1.0, "k", "-", 0.0)]
                for half_sign, half_color, half_style, half_order in halves:
                    half_cap = (cap_center[np.newaxis, :]
                           + half_sign * boundary_offset_normal * np.cos(0.5 * phase)[:, np.newaxis] * radius
                           + boundary_offset_direction * np.sin(0.5 * phase)[:, np.newaxis] * radius)
                    projection.plot_curve(half_cap, color=half_color, linestyle=half_style, zorder=half_order)
            bottom_line.append(cap_center - boundary_offset_direction * radius)
            top_line.append(cap_center + boundary_offset_direction * radius)
        projection.plot_curve(np.asarray(bottom_line), color="k")
        projection.plot_curve(np.asarray(top_line), color="k")
        # Shade the cylinder
        for cap in cap_list:
            cap_boundary = projection.transform(cap)
            pyplot.fill(cap_boundary[:, 0], cap_boundary[:, 1], color=light_color, zorder=-3000.0)
        rectangle = np.asarray(bottom_line + top_line[::-1])
        rectangle_boundary = projection.transform(rectangle)
        pyplot.fill(rectangle_boundary[:, 0], rectangle_boundary[:, 1], color=light_color, zorder=-3000.0)
    # Annotate a few points
    closest_point_on_line = closest_point * min_distance
    point_list = []
    if "endpoints" in elements:
        point_list += [(line_begin, r"$l_0$"), (line_end, r"$l_1$")]
    if "closest_point" in elements:
        point_list += [(closest_point_on_line, r"$q$")]
    for point, label in point_list:
        point_x, point_y = projection.transform(point)[0:2]
        plot_annotated_point(point_x, point_y, label, "left", "bottom", [0.017, 0.0])
    # Annotate the radius
    if "radius" in elements:
        radius_line = np.asarray([line_begin, line_begin + boundary_offset_direction * radius])
        projection.plot_curve(radius_line, color="k")
        center_x, center_y = np.mean(projection.transform(radius_line)[:, 0:2], axis=0)
        pyplot.text(center_x + 0.03, center_y, r"$R$")
    if "extent_t" in elements:
        plot_annotated_length(True, min_distance - radius, min_distance + radius, 0.83, r"$2R$")
    # Annotate the opening angle
    if "triangle_boundary" in elements:
        silhouette_point = np.cos(opening_angle) * closest_point + np.sin(opening_angle) * line_normal
        silhouette_point *= np.sqrt(min_distance ** 2 - radius ** 2)
        triangle = [closest_point_on_line, silhouette_point, np.zeros(3)]
        plot_right_triangle(projection, triangle, r"", r"$R$", "", 0.66, r"$\varphi_R$", 0.1, (0.09, 0.09, 0.09))
    # Annotate the extent along the s-axis
    if "extent_s" in elements:
        for line_point in [line_begin, line_end]:
            to_line = np.asarray([np.zeros(3), line_point])
            projection.plot_curve(to_line, color="k", linestyle=":")
            plot_annotated_length(False, line.begin_s, line.begin_s + line.extent_s, 1.015,
                                  "", r"$l_{0,s}$", r"$l_{1,s}$", dict(), 0.06)
    # Plot the construction of the closest point
    if "triangle_end" in elements:
        central_axis = np.asarray([line_begin, line_end])
        projection.plot_curve(central_axis, linestyle=":", color=0.8 * light_color)
        triangle = np.asarray([np.zeros(3), closest_point_on_line, line_end])
        plot_right_triangle(projection, triangle, "", "", "")
    # Plot the coordinate frame constructed for the cylinder
    axes_length = 1.0
    if len(frozenset(["axes_s", "axes_st", "axes_su", "axes_stu"]) & frozenset(elements)):
        plot_axis(projection, axes_origin, line_direction * axes_length, r"$s$")
    if len(frozenset(["axes_t", "axes_st", "axes_tu", "axes_stu"]) & frozenset(elements)):
        plot_axis(projection, axes_origin, closest_point * axes_length, r"$t$", True)
    if len(frozenset(["axes_u", "axes_su", "axes_tu", "axes_stu"]) & frozenset(elements)):
        plot_axis(projection, axes_origin, line_normal * axes_length, r"$u$", True, 0.15)
    if "solid_angle" in elements:
        # Form a rectangle approximating the solid angle of the cylinder
        vertices = list()
        for i, coords in enumerate([(1.0, 1.0), (0.0, 1.0), (0.0, -1.0), (1.0, -1.0)]):
            s = line.begin_s + coords[0] * line.extent_s
            tu = np.sqrt(1.0 - s * s)
            t = np.cos(opening_angle) * tu
            u = coords[1] * np.sin(opening_angle) * tu
            vertices.append(s * line_direction + t * closest_point + u * line_normal)
        vertices = np.asarray(vertices)
        # Draw the outline of the solid angle
        outline_curve = sample_edges(vertices, True)
        projection.plot_curve(outline_curve, color=solid_angle_line_color, zorder=-1.0)
        # Handle front- and backfacing parts separately for the fill
        for depth_factor in [-1.0, 1.0]:
            # Transform and clip to find the relevant parts
            screen_vertices = projection.transform(vertices)
            screen_vertices[..., 2] *= depth_factor
            new_vertex_count, screen_vertices = clip_polygon(screen_vertices, screen_vertices.shape[0] + 1)
            if new_vertex_count > 0:
                screen_vertices = screen_vertices[0:new_vertex_count, :]
                outline = sample_edges(screen_vertices, True)
                pyplot.fill(outline[:, 0], outline[:, 1], zorder=-5000.0, color=solid_angle_color)

    rcParams["font.size"] = previous_size


# These images are used by several slides so it is best to load them just once
motivation_result_images = [
    imread(path.join("../vulkan_renderer/data/experiments",
        [name for name in listdir("../vulkan_renderer/data/experiments") if re.match(r"cornell_box_%s_.*\.png" % technique, name)][0]))
    for technique in ["area", "solid_angle", "projected_solid_angle"]
]


class RendererSlide(Slide):
    """Starts my renderer for a live demo."""

    def __init__(self, experiment=-1, show_gui=False, v_sync=True, title=None, image=None):
        super().__init__()
        self.process = None
        self.title = title
        if image is not None:
            self.image = imread(image)
        else:
            self.image = None
        self.args = [
            "-e%d" % experiment,
            "-gui" if show_gui else "-no_gui",
            "-v_sync" if v_sync else "-no_v_sync",
        ]

    def animate(self):
        if self.process is not None and self.process.poll() is not None:
            self.process = None
            return "step_forward"

    def on_step(self, state, previous_state):
        # Run the renderer when we move forward through the presentation
        if state[0] > previous_state[0]:
            self.process = Popen(["../vulkan_renderer/build/vulkan_renderer"] + self.args, cwd="../vulkan_renderer")
        # Just skip this slide when moving backward
        else:
            return "step_backward"

    def draw(self):
        if self.title is not None:
            draw_title(self.title)
        if self.image is not None:
            draw_image(self.image, (0.0, 0.0), 16.0)


class CylinderSlide(Slide):
    """A slide with a big cylindrical light. Has several variants."""

    def __init__(self, variant):
        """
        Prepares drawing.
        :param variant: Can be one of the following strings to select different
            variants of the slide.
            - linear_light: Definition of linear lights as thin cylinders.
            - solid_angle: Illustrate computation of the solid angle.
        """
        super().__init__()
        self.yaw, self.pitch = -0.17 * np.pi, 0.03 * np.pi
        self.yaw += np.pi / 180.0
        self.line_axes = None
        # Define the line and the cylinder
        self.line_begin = np.asarray([-1.3, -0.3, 1.8])
        self.line_end = np.asarray([0.9, 0.1, 1.5])
        self.radius = 0.4
        self.variant = variant
        if variant == "linear_light":
            self.step_count = 2
            self.step_durations = [3.0]
        elif variant == "solid_angle":
            self.step_count = 10
            self.step_durations = [0.0, 0.0, 0.0, 2.0, 0.0, 2.0, 0.0, 0.0, 0.0, 2.0]
        else:
            self.step_count = 4
            self.step_durations = [2.0]

    def draw(self):
        if self.variant == "linear_light":
            draw_title(r"Linear light $\coloneq$ infinitesimally thin cylinder")
            pyplot.text(8.0, 2.0, r"$R\rightarrow0$", ha="center", va="bottom", fontsize=36.0)
        elif self.variant == "solid_angle":
            draw_title("The solid angle of a cylinder")
        else:
            draw_title("The projected solid angle of a line")
        slide_axes = pyplot.gca()
        self.line_axes = slide_axes.inset_axes((0.0, 0.0, 15.0, 8.0), transform=slide_axes.transData, aspect="equal", frameon=False)
        pyplot.gcf().add_axes(self.line_axes)
        self.draw_line()
        pyplot.sca(slide_axes)
        # Draw some bulleted text
        psa_x, psa_y = 7.6, 5.8
        long_solid_angle_formula = r"$\Omega_R \,\coloneq\, \int_{l_{0,s}}^{l_{1,s}} \int_{-\varphi_R}^{\varphi_R} 1 \,\mathrm{d}\varphi \,\mathrm{d}\omega_s$"
        solid_angle_formula = r"$\Omega_R \,\coloneq\, 2\varphi_R \int_{l_{0,s}}^{l_{1,s}} 1 \,\mathrm{d}\omega_s$"
        if self.variant == "solid_angle":
            bullets = [
                (0.9, r"$\varphi_R \,\coloneq\, \arcsin\frac{R}{\|q\|}$"),
                (0.9, r"$l_{0,s} \,\coloneq\, \frac{l_0\cdot s}{\|l_0\|}$,   $l_{1,s} \,\coloneq\, \frac{l_1\cdot s}{\|l_1\|}$"),
                (1.6, long_solid_angle_formula if self.step < 9 else solid_angle_formula),
                (1.1, r"$\frac{\mathrm{d}\Omega_R}{\mathrm{d}R} \,\coloneq\, \lim_{R\rightarrow 0} \frac{\Omega_R}{R} = 2\,\frac{l_{1,s}-l_{0,s}}{\|q\|}$"),
                (1.6, r"$\frac{\mathrm{d}\Omega_R^\bot}{\mathrm{d}R} \,\coloneq\, \frac{2}{\|q\|}\int_{l_{0,s}}^{l_{1,s}} n\cdot(\omega_s s + \sqrt{1-\omega_s^2} t) \,\mathrm{d}\omega_s$"),
            ][0:[0, 0, 0, 0, 0, 1, 1, 2, 3, 3][self.step]]
            x, y = 10.2, 7.7
            for i, bullet in enumerate(bullets):
                line_height, text = bullet
                y -= line_height
                if self.step == 9 and i == 2:
                    x = smooth_lerp(x, psa_x, self.step_lerp)
                    y = smooth_lerp(y, psa_y, self.step_lerp)
                if i >= 2 or self.step < 9 or (self.step == 9 and self.step_lerp == 0.0):
                    pyplot.text(x, y, text, ha="left", va="bottom")
        elif self.variant == "projected_solid_angle":
            bullets = [
                (0.0, solid_angle_formula),
                (1.2, r"$\frac{\mathrm{d}\Omega_R}{\mathrm{d}R} \,\coloneq\, \lim_{R\rightarrow 0} \frac{\Omega_R}{R} = \frac{2}{\|q\|} \int_{l_{0,s}}^{l_{1,s}} 1 \,\mathrm{d}\omega_s$"),
                (1.4, r"$\frac{\mathrm{d}\Omega_R^\bot}{\mathrm{d}R} \,\coloneq\, \frac{2}{\|q\|}\int_{l_{0,s}}^{l_{1,s}} n\cdot(\omega_s s + \sqrt{1-\omega_s^2} t) \,\mathrm{d}\omega_s$"),
                (1.6, r"$\,~~~~=\, \frac{2}{\|q\|}\int_{l_{0,s}}^{l_{1,s}} n_s \omega_s + n_t \sqrt{1-\omega_s^2} \,\mathrm{d}\omega_s$"),
            ][0:[1, 2, 3, 4][self.step]]
            x, y = psa_x, psa_y
            for i, bullet in enumerate(bullets):
                line_height, text = bullet
                y -= line_height
                pyplot.text(x, y, text, ha="left", va="bottom")

    def on_mouse_move(self, position, motion, held_buttons):
        if "lmb" in held_buttons:
            self.yaw += motion[0] * 0.003
            self.pitch += motion[1] * 0.003
            self.draw_line()
            return "redraw_canvas"

    def animate(self):
        if self.variant == "solid_angle" and self.step == 9 and self.lerp_ongoing:
            return "draw"
        elif self.lerp_ongoing:
            self.draw_line()
            return "redraw_canvas"

    def draw_line(self):
        self.line_axes.clear()
        if self.variant == "linear_light":
            self.line_axes.set_xlim(-3.5, 2.5)
        else:
            limit_lerp = self.step_lerp if self.step == 9 else (0.0 if self.step < 9 else 1.0)
            if self.variant == "projected_solid_angle":
                limit_lerp = 1.0
            self.line_axes.set_xlim(smooth_lerp(-1.65, -1.58, limit_lerp), 3.0)
        self.line_axes.set_ylim(-0.5, 2.3)
        self.line_axes.set_xticks([])
        self.line_axes.set_yticks([])
        pyplot.sca(self.line_axes)
        # Prepare rotations for various views
        line_dir = (self.line_end - self.line_begin) / np.linalg.norm(self.line_end - self.line_begin)
        closest_dir = get_closest_point(self.line_begin, self.line_end)
        line_normal = np.cross(line_dir, closest_dir)
        user_rot = get_euler_projection(self.yaw, self.pitch).matrix
        side_rot = [line_dir, closest_dir, line_normal]
        cap_rot = [line_normal, closest_dir, -line_dir]
        axes_origin = (1.2, -0.3, 0.15)
        if self.variant == "linear_light":
            if self.step == 0:
                radius = smooth_lerp(0.4, 0.0, self.step_lerp)
            else:
                radius = 0.0
            axes_origin = (0.0, 0.0, 0.0)
            projection = get_euler_projection(self.yaw, self.pitch)
            elements = ["cylinder", "radius"]
        elif self.variant == "solid_angle":
            # Define what will be shown in each step
            step_state = [
                (user_rot, ["hemisphere", "solid_angle", "cylinder"]),
                (user_rot, ["hemisphere", "solid_angle", "cylinder", "endpoints"]),
                (user_rot, ["hemisphere", "solid_angle", "cylinder", "endpoints", "closest_point", "triangle_end"]),
                (user_rot, ["hemisphere", "solid_angle", "cylinder", "endpoints", "closest_point", "triangle_end", "axes_stu"]),
                (cap_rot, ["hemisphere", "solid_angle", "cylinder", "axes_stu", "closest_point"]),
                (cap_rot, ["hemisphere", "solid_angle", "cylinder", "axes_stu", "triangle_boundary", "closest_point"]),
                (side_rot, ["hemisphere", "solid_angle", "cylinder", "endpoints", "closest_point", "axes_stu"]),
                (side_rot, ["hemisphere", "solid_angle", "cylinder", "endpoints", "closest_point", "axes_stu", "extent_s"]),
            ]
            clamped_step = np.minimum(len(step_state) - 1, self.step)
            elements = step_state[clamped_step][1]
            # Do not show certain elements during an animation
            elements = [element for element in elements if element not in ["extent_s", "triangle_end", "triangle_boundary"] or self.step_lerp == 0.0]
            # Interpolate between the relevant rotations
            slerp = spatial.transform.rotation.Slerp([0.0, 1.0], spatial.transform.Rotation.from_matrix([step_state[clamped_step][0], step_state[np.minimum(len(step_state) - 1, self.step + 1)][0]]))
            projection = OrthogonalProjection(slerp(smooth_lerp(0.0, 1.0, self.step_lerp)).as_matrix())
            radius = self.radius
        else:
            axes_origin = (-1.2, -0.3, -0.1)
            projection = OrthogonalProjection(side_rot)
            elements = ["hemisphere", "solid_angle", "cylinder"]
            if self.step >= 1:
                elements += ["closest_point"]
            if self.step >= 2:
                elements += ["axes_st"]
                # Display the normal vector
                transformed_normal = projection.transform(np.asarray([0.0, 0.0, 1.0]))
                length = 0.88
                pyplot.arrow(0.0, 0.0, length * transformed_normal[0], length * transformed_normal[1], head_width=0.08, facecolor="w")
                label_position = transformed_normal * 0.6
                pyplot.text(label_position[0], label_position[1], "$n$", ha="left", va="bottom")
            # Animate the radius
            radius_lerp = self.step_lerp if self.step == 0 else 1.0
            radius = smooth_lerp(self.radius, 0.0, radius_lerp)
        plot_line_solid_angle_talk(elements, self.line_begin, self.line_end, radius, projection, axes_origin)


class GamitoSlide(Slide):
    """Illustrates how Gamito's approach could be applied."""

    def __init__(self):
        super().__init__()
        self.gamito_image = imread("../Graphics/line_talk/gamito_rectangle.png")
        self.peters_image = imread("../Graphics/line_talk/projected_solid_angle.png")

    def draw(self):
        draw_title("BRDF importance sampling for cylindrical lights")
        draw_image(self.gamito_image, (2.0, 0.6), 5.0)
        draw_image(self.peters_image, (8.5, 0.6), 6.0)
        pyplot.text(8.0 - 3.5, 7.0, "Gamito 2016", ha="center", va="bottom")
        pyplot.text(8.0, 7.0, "+", ha="center", va="bottom")
        pyplot.text(8.0 + 3.5, 7.0, "Peters 2021", ha="center", va="bottom")


class MotivationSlide(Slide):

    def __init__(self, measure):
        """
        A slide to illustrate benefits of different sampling strategies.
        :param measure: "area", "solid" or "projected".
        """
        super().__init__()
        self.yaw, self.pitch = -2.284, 0.154
        self.line_axes = None
        self.measure = measure
        # Define the line
        self.line_begin = np.asarray([-0.4, 2.3, 1.8])
        self.line_end = np.asarray([0.9, 1.4, 0.1])
        self.line_length = np.linalg.norm(self.line_end - self.line_begin)
        self.line_direction = (self.line_end - self.line_begin) / self.line_length
        # Create samples uniformly with respect to the appropriate measure
        random_numbers = get_random_numbers()
        self.samples = np.zeros((random_numbers.shape[0], 3))
        if self.measure == "area":
            self.step_count = 3
            for j in range(random_numbers.shape[0]):
                self.samples[j] = self.line_begin + random_numbers[j] * self.line_length * self.line_direction
            self.samples /= np.linalg.norm(self.samples, axis=1)[:, np.newaxis]
        elif self.measure == "solid":
            prepared_line = prepare_solid_angle_line_sampling(self.line_begin, self.line_direction, self.line_length)
            for j in range(random_numbers.shape[0]):
                self.samples[j] = sample_solid_angle_line(prepared_line, random_numbers[j])
        elif self.measure == "projected":
            prepared_line = prepare_projected_solid_angle_line_sampling_without_clipping(np.asarray([0.0, 0.0, 1.0]), self.line_begin, self.line_direction, self.line_length)
            for j in range(random_numbers.shape[0]):
                self.samples[j] = sample_projected_solid_angle_line(prepared_line, random_numbers[j])
        elif self.measure == "ltc":
            shading_pitch = 0.5 * np.pi
            world_to_shading_space = np.asarray([
                [np.cos(shading_pitch), -np.sin(shading_pitch), 0.0],
                [np.sin(shading_pitch), np.cos(shading_pitch), 0.0],
                [0.0, 0.0, 1.0],
            ])
            # shading_to_cosine_space = np.asarray([[0.733347774, 0.000000, -0.0798025057], [0.000000, 0.786373079, 0.000000], [0.113918357, 0.000000, 1.0]]).T
            # shading_to_cosine_space = np.asarray([[0.00990602188, 0.000000, -0.00124329492], [0.000000, 1.0, 0.000000], [0.0985978469, 0.000000, 0.000127096064]]).T
            # shading_to_cosine_space = np.asarray([[0.0105073005, 0.000000, -0.00500859041], [0.000000, 1.0, 0.000000], [0.102492839, 0.000000, 0.000652103685]]).T
            # shading_to_cosine_space = np.asarray([[0.0309120584, 0.000000, -0.0297419745], [0.000000, 1.0, 0.000000], [0.183499739, 0.000000, 0.0111321807]]).T
            shading_to_cosine_space = np.asarray([[0.117774136, 0.000000, -0.105238415], [0.000000, 1.0, 0.000000], [0.319090545, 0.000000, 0.0845179483]]).T
            world_to_cosine_space = shading_to_cosine_space @ world_to_shading_space
            line_begin = self.line_begin @ world_to_cosine_space.T
            line_end = self.line_end @ world_to_cosine_space.T
            line_length = np.linalg.norm(line_end - line_begin)
            prepared_line = prepare_projected_solid_angle_line_sampling_without_clipping(np.asarray([0.0, 0.0, 1.0]), line_begin, (line_end - line_begin) / line_length, line_length)
            for j in range(random_numbers.shape[0]):
                self.samples[j] = sample_projected_solid_angle_line(prepared_line, random_numbers[j])
            self.samples = self.samples @ np.linalg.inv(world_to_cosine_space).T
            self.samples /= np.linalg.norm(self.samples, axis=-1)[..., np.newaxis]
        # Project them onto the light source
        self.light_normal = get_closest_point(self.line_begin, self.line_end)
        light_offset = self.light_normal @ self.line_begin
        self.light_samples = self.samples * light_offset / np.einsum("...i, i -> ...", self.samples, self.light_normal)[:, np.newaxis]

    def draw(self):
        title_dict = dict(area="Area sampling", solid="Solid angle sampling",
                          projected="Projected solid angle sampling (ours)", ltc="LTC importance sampling")
        draw_title(title_dict[self.measure])
        text_x = 7.6
        rendering_equation = r"$L_o(\omega_o)=     \int_\Omega L_i(\omega_i) f_r(\omega_i, \omega_o) \, n\cdot\omega_i \,\mathrm{d}\omega_i$"
        if self.measure == "area":
            pyplot.text(text_x, 5.0, rendering_equation)
            if self.step >= 2:
                pyplot.text(text_x, 3.4, r"$\,~~~~~p(\omega_i) \propto L_e(\omega_i) \, \frac{\|x-y\|^2}{n_l \cdot \omega_i}$")
        elif self.measure == "solid":
            pyplot.text(text_x, 5.0, rendering_equation)
            pyplot.text(text_x, 3.5, r"$\,~~~~~p(\omega_i) \propto L_e(\omega_i)$")
        elif self.measure == "projected":
            pyplot.text(text_x, 5.0, rendering_equation)
            pyplot.text(text_x, 3.5, r"$\,~~~~~p(\omega_i) \propto L_e(\omega_i) \, n\cdot\omega_i$")
        elif self.measure == "ltc":
            pyplot.text(text_x, 5.0, rendering_equation)
            pyplot.text(text_x, 3.5, r"$\,~~~~~p(\omega_i) \propto L_e(\omega_i) f_r(\omega_i, \omega_o) \, n\cdot\omega_i$")
        # Draw a fraction bar below the integrand of the rendering equation
        if self.measure != "area" or self.step >= 2:
            pyplot.plot([10.1, 14.4], 2 * [4.3], color="k")
        slide_axes = pyplot.gca()
        self.line_axes = slide_axes.inset_axes((0.3, 0.7, 7.1, 7.1), transform=slide_axes.transData,
                                                  aspect="equal", frameon=False)
        pyplot.gcf().add_axes(self.line_axes)
        self.draw_line()

    def on_mouse_move(self, position, motion, held_buttons):
        if "lmb" in held_buttons:
            self.yaw += motion[0] * 0.003
            self.pitch += motion[1] * 0.003
            self.draw_line()
            return "redraw_canvas"

    def draw_line(self):
        self.line_axes.clear()
        self.line_axes.set_xlim(-1.05, 2.8)
        self.line_axes.set_ylim(-0.85, 3.0)
        self.line_axes.set_xticks([])
        self.line_axes.set_yticks([])
        pyplot.sca(self.line_axes)
        pyplot.scatter([0.0], [0.0], color="k", s=4.0)
        elements_dict = dict(
            area=(("cylinder", "hemisphere") if self.step == 0 else ("cylinder", "hemisphere", "solid_angle")),
            solid=("cylinder", "hemisphere", "solid_angle"),
            projected=("cylinder", "hemisphere", "solid_angle"),
            ltc=("cylinder", "hemisphere", "solid_angle"),
        )
        projection = get_euler_projection(self.yaw, self.pitch)
        plot_line_solid_angle_talk(elements_dict[self.measure], self.line_begin, self.line_end, 0.0, projection)
        # Plot normal vectors
        origins = [np.asarray(3 * [0.0]), 0.5 * (self.line_begin + self.line_end)]
        normals = [np.asarray([0.0, 0.0, 1.0]), self.light_normal]
        normal_count = 2 if self.measure == "area" and self.step >= 2 else 1
        for label, origin, normal in list(zip(("$n$", "$n_l$"), origins, normals))[0:normal_count]:
            transformed_origin = projection.transform(origin)
            transformed_normal = projection.transform(normal)
            length = 0.88
            pyplot.arrow(transformed_origin[0], transformed_origin[1], length * transformed_normal[0], length * transformed_normal[1],
                         head_width=0.08, facecolor="w")
            label_position = transformed_origin + transformed_normal * 0.6
            pyplot.text(label_position[0], label_position[1], label, ha="left", va="bottom")
        # Label the solid angle of the line
        # if self.measure == "area" and self.step >= 1:
        #     line_label_position = projection.transform(origins[1] / np.linalg.norm(origins[1]))
        #     pyplot.text(line_label_position[0], line_label_position[1], r"$\mathrm{\mathbb{L}}$", ha="right", va="bottom")
        # Draw the samples
        samples_list = list()
        if self.measure == "area":
            if self.step >= 1:
                samples_list = [("C1", self.light_samples)]
            if self.step >= 2:
                samples_list += [("k", self.samples)]
        else:
            samples_list = [("k", self.samples)]
        for color, samples in samples_list:
            transformed_samples = projection.transform(samples)
            pyplot.scatter(transformed_samples[:, 0], transformed_samples[:, 1], s=1.0, c=color, zorder=1000.0)


class MotivationResultSlide(ImageComparisonSlide):
    """Shows Cornell box renderings."""

    def __init__(self, image_count=3):
        super().__init__()
        self.images = motivation_result_images[0:image_count]
        self.labels = ["Area", "Solid angle", "Projected solid angle"]
        self.left_bottom = 0.35, 1.2
        self.image_width = 5.0
        self.offset = 5.1, 0.0

    def draw(self):
        super().draw()
        draw_title(("Results" if len(self.images) > 1 else "Result") + " at 1 sample per pixel")


class TzuMaoLiSlide(Slide):
    """Illustrates why we do not want Tzu-Mao Li's solution."""

    def __init__(self):
        super().__init__()
        self.code_image = imread("../Graphics/line_talk/tzu_mao_li_code.png")
        self.illustrative_image = imread("../Graphics/line_talk/tzu_mao_li_illustrative.jpg")

    def draw(self):
        draw_title("Projected solid angle sampling [Li et al. 2018]")
        draw_image(self.code_image, (2.0, 0.6), 12.0)
        draw_image(self.illustrative_image, (5.2, 4.7), 5.6)


# We use globals so that interactive state is preserved across slides
cdf_slide_local_normal = np.asarray([np.cos(0.9), np.sin(0.9)])
cdf_slide_target_area = 0.4


def solve_symmetric_homogeneous_quadratic(quadratic):
    """
    Solves a quadratic equation given by a symmetric 2x2 matrix (assuming that
    a real solution exists).
    :return: A 2-vector such that vector.T @ quadratic @ vector == 0.0. Pass
        -quadratic to get the other root.
    """
    coeff_xy = quadratic[0, 1]
    sqrt_discriminant = np.sqrt(np.maximum(0.0, coeff_xy * coeff_xy - quadratic[0, 0] * quadratic[1, 1]))
    scaled_root = abs(coeff_xy) + sqrt_discriminant
    return np.asarray([scaled_root, -quadratic[0, 0]]) if coeff_xy >= 0.0 else np.asarray([quadratic[1, 1], scaled_root])


def line_sampling_iteration(normal, angle, cdf):
    """
    Performs a single iteration of the projected solid angle sampling procedure
    for lines.
    :param normal: The s- and t-coordinate of the shading normal.
    :param angle: The arcsin of the s-coordinate of the normalized current guess
        for the sample.
    :param cdf: The target value for the CDF.
    :return: The angle after the iteration and coefficients of the quadratic
        Taylor expansion at the given angle (ordered by increasing exponent).
    """
    dir = np.asarray([np.sin(angle), np.cos(angle)])
    # Evaluate the objective function and two derivatives
    dot_0 = normal @ dir
    objective = dot_0 * dir[0] + (angle * normal[1] - cdf)
    half_derivative_1 = dot_0 * dir[1]
    dot_1 = normal @ np.asarray([dir[1], -dir[0]])
    half_derivative_2 = dot_1 * dir[1] - dot_0 * dir[0]
    # Compute the root of minimal magnitude for a quadratic Taylor
    # expansion. This is Laguerre's method for n = 2, also known as
    # Halley's irrational formula.
    root = solve_symmetric_homogeneous_quadratic(-np.asarray([[half_derivative_2, half_derivative_1], [half_derivative_1, objective]]))
    # On to the next iteration, or maybe we are done
    new_angle = angle + root[0] / root[1]
    new_angle = np.clip(new_angle, -0.5 * np.pi, 0.5 * np.pi)
    # Compute the random number that the input corresponds to to figure out if
    # we should iterate at all
    min_s, max_s = -normal[1], 1.0
    random_number = (np.sin(angle) - min_s) / (max_s - min_s)
    if 1.0e-5 < random_number < 1.0 - 1.0e-5:
        angle = new_angle
    return angle, np.asarray([objective + cdf, 2.0 * half_derivative_1, half_derivative_2])


class CDFSlide(Slide):
    """Illustrates the distribution function that needs to be inverted."""

    def __init__(self, variant):
        """
        :param variant: Either "s" to use the s-coordinate of the sample for the
            x-axis or "angle" to use the angle or "newton" to use the angle but
            Newton's method instead of ours.
        """
        super().__init__()
        self.variant = variant
        self.circle_center = np.asarray([13.0, 3.2])
        self.circle_radius = 2.0
        self.slide_axes = None
        self.graph_axes = None
        self.step_count = dict(s=5, angle=3, newton=1)[self.variant]

    def on_click(self, position, button):
        return self.on_mouse_move(position, (0.0, 0.0), frozenset([button]))

    def on_mouse_move(self, position, motion, held_buttons):
        global cdf_slide_local_normal, cdf_slide_target_area
        if "lmb" in held_buttons and self.slide_axes is not None:
            position_data = self.slide_axes.transData.inverted().transform(position)
            circle_offset = np.asarray(position_data) - self.circle_center
            if circle_offset[0] > - self.circle_radius:
                circle_offset[0] = np.maximum(0.0, circle_offset[0])
                cdf_slide_local_normal = circle_offset / np.linalg.norm(circle_offset)
                return "draw"
            elif self.graph_axes is not None:
                position_graph = self.graph_axes.transData.inverted().transform(position)
                cdf_slide_target_area = position_graph[1]
                return "draw"

    def draw(self):
        self.slide_axes = pyplot.gca()
        if self.variant == "newton":
            draw_title("Newton's method")
        else:
            draw_title("Cumulative distribution function")
        if self.variant == "s":
            draw_bullets([
                r"$F_{n_s,n_t}(\omega_s) \,\coloneq\, 2\int_0^{\omega_s} n_s \omega_s^\prime + n_t \sqrt{1-\omega_s^{\prime2}} \,\mathrm{d}\omega_s^\prime$",
                r"$~~~~~~~~~~~~~~\, = \,~~~~~~~ n_s \omega_s^2 + n_t (\sqrt{1-\omega_s^2} \omega_s + \mathrm{arcsin}\, \omega_s)$",
            ][0:[1, 2, 2, 2, 2][self.step]], 1.1, 6.8)
        else:
            draw_bullets([
                r"$\alpha \,\coloneq\, \arcsin\omega_s$",
                r"$G_{n_s,n_t}(\alpha) \,\coloneq\, F_{n_s,n_t}(\sin\alpha) = n_s \sin^2\alpha + n_t (\cos\alpha \, \sin\alpha + \alpha)$",
            ], 0.9, 6.8)
        if self.step >= 2 or self.variant != "s":
            # Plot the interactive normal picker
            center = self.circle_center
            n = cdf_slide_local_normal
            arrow_length = 0.86 * self.circle_radius
            phases = np.linspace(-np.pi, np.pi, 513)
            pyplot.plot(center[0] + self.circle_radius * np.cos(phases), center[1] + self.circle_radius * np.sin(phases), color=horizon_color)
            pyplot.arrow(center[0], center[1], arrow_length * n[0], arrow_length * n[1], head_width=0.16, facecolor="w", zorder=1000.0)
            text_position = center + 0.5 * self.circle_radius * np.asarray(n) - 0.23 * np.asarray([-n[1], n[0]])
            pyplot.text(text_position[0], text_position[1], r"$n$", ha="center", va="center")
            # Plot the s- and t-axis
            plot_line_solid_angle_talk(["axes_st"], np.asarray([0.0, 1.0, 0.0]), np.asarray([1.0, 1.0, 0.0]), 0.0, get_euler_projection(0.0, 0.5 * np.pi), np.concatenate([center - self.circle_radius, [0.0]]))
            # Prepare axes and values for the graph
            self.graph_axes = self.slide_axes.inset_axes((2.0, 1.5, 7.5, 3.5), transform=self.slide_axes.transData)
            s = np.linspace(-n[1], 1.0, 2049)
            F = n[0] * s * s + n[1] * (np.sqrt(1.0 - s * s) * s + np.arcsin(s))
            # Prepare the initialization
            target_area = np.clip(cdf_slide_target_area, np.min(F), np.max(F))
            init_F = np.interp(s, [s[0], s[-1]], [F[0], F[-1]])
            init_s = np.interp(target_area, [F[0], F[-1]], [s[0], s[-1]])
            # Prepare the ground truth
            final_angle = np.arcsin(init_s)
            for i in range(2):
                final_angle = line_sampling_iteration(n, final_angle, target_area)[0]
            final_s = np.sin(final_angle)
            # Plot the function
            radian_ticks = [-0.5 * np.pi, 0.0, 0.5 * np.pi]
            radian_labels = [r"$-\frac{\pi}{2}$", r"0", r"$\frac{\pi}{2}$"]
            self.graph_axes.set_ylim(-0.5 * np.pi - 0.1, 0.5 * np.pi + 0.4)
            axes = self.graph_axes
            if self.step >= 3 or self.variant != "s":
                axes.set_yticks(radian_ticks + [target_area])
                axes.set_yticklabels(radian_labels + [r""])
                axes.plot([-10.0, 10.0], 2 * [target_area], ":", color="k", linewidth=0.5, zorder=100.0)
            else:
                axes.set_yticks(radian_ticks)
                axes.set_yticklabels(radian_labels)
            if self.variant == "s":
                axes.set_xlabel(r"$\omega_s$")
                axes.set_xlim(-1.05, 1.05)
                axes.plot(s, F, label=r"$F_{n_s,n_t}$")
                if self.step == 3:
                    axes.scatter([final_s], [target_area], color="k", zorder=200.0, s=4.0)
                if self.step >= 4:
                    axes.plot(s, init_F, label="Init.")
                    axes.scatter([init_s], [target_area], color="k", zorder=200.0, s=4.0)
            else:
                angle = np.arcsin(s)
                axes.set_xlabel(r"$\alpha$")
                axes.set_xlim(-0.53 * np.pi, 0.53 * np.pi)
                axes.set_xticks(radian_ticks)
                axes.set_xticklabels(radian_labels)
                axes.plot(angle, F, label=r"$G_{n_s,n_t}$")
                axes.plot(angle, init_F, label="Init.")
                # Perform the appropriate number of iterations
                iteration = [0, 0, 1, 0][self.step]
                next_angle = current_angle = np.arcsin(init_s)
                for j in range(iteration + 1):
                    current_angle = next_angle
                    next_angle, parabola = line_sampling_iteration(n, current_angle, target_area)
                axes.scatter([current_angle], [target_area], color="k", zorder=200.0, s=4.0)
                if self.step >= 1 or self.variant == "newton":
                    axes.scatter([current_angle], [parabola[0]], color="k", zorder=200.0, s=4.0)
                    axes.plot(2 * [current_angle], [-10.0, 10.0], ":", color="k", linewidth=0.5, zorder=100.0)
                    # Plot a parabola and the next point
                    if self.step != 3 and self.variant == "angle":
                        axes.plot(angle, np.polyval(parabola[::-1], (angle - current_angle)), color="k", linewidth=1.4)
                        axes.plot(angle, np.polyval(parabola[::-1], (angle - current_angle)), color="w", linewidth=0.6)
                        axes.scatter([next_angle], [target_area], color="w", ec="k", zorder=200.0, s=4.0)
                    # Plot a tangent to illustrate where Newton would go
                    else:
                        axes.plot(angle, np.polyval(parabola[::-1][1:], (angle - current_angle)), color="k", linewidth=1.4)
                        axes.plot(angle, np.polyval(parabola[::-1][1:], (angle - current_angle)), color="w", linewidth=0.6)
                        next_angle_newton = current_angle - (parabola[0] - target_area) / parabola[1]
                        axes.scatter([next_angle_newton], [target_area], color="w", ec="k", zorder=200.0, s=4.0)
            axes.legend(frameon=False, loc="upper left", handlelength=1.0, handletextpad=0.5, borderpad=0.0)


class LTCResultsSlide(ImageComparisonSlide):
    """Shows renderings that motivate the need for LTC sampling."""

    def __init__(self, use_ltc):
        super().__init__()
        experiment_dir = "../vulkan_renderer/data/experiments"
        suffix = "clamped_optimal_mis" if use_ltc else "projected_solid_angle"
        image_name = [name for name in listdir(experiment_dir) if re.match(r"bistro_outside_%s_.*\.png" % suffix, name)][0]
        self.images = [imread(path.join(experiment_dir, image_name))]
        self.labels = ["MIS: LTC + projected solid angle sampling" if use_ltc else "Projected solid angle sampling"]
        self.left_bottom = 1.8, 0.2
        self.image_width = 12.4
        self.offset = 5.1, 0.0
        self.half_magnifier_extent = 24, 24
        self.spp = 2 if use_ltc else 1

    def draw(self):
        super().draw()
        draw_title("Diffuse and specular shading (%d spp)" % self.spp)


class LTCSlide(Slide):
    """Visualizes sampling of LTCs."""

    def __init__(self, title):
        super().__init__()
        self.title = title
        self.step_count = 2
        self.slide_axes = self.line_axes = None
        self.yaw, self.pitch = -0.1 * np.pi, 0.08 * np.pi
        self.projection = get_euler_projection(self.yaw, self.pitch)
        shading_pitch = -0.6 * np.pi
        self.world_to_shading_space = np.asarray([
            [np.cos(shading_pitch), -np.sin(shading_pitch), 0.0],
            [np.sin(shading_pitch), np.cos(shading_pitch), 0.0],
            [0.0, 0.0, 1.0],
        ])
        self.shading_to_cosine_space = np.asarray([[0.782347, 0.000000, -0.134090], [0.000000, 1.000000, 0.000000], [0.422520, 0.000000, 0.262691]]).T
        self.world_to_cosine_space = self.shading_to_cosine_space @ self.world_to_shading_space
        self.line_world_space = np.asarray([[-1.3, -2.6, 1.8], [0.9, -0.5, 1.5]])
        self.line_world_space /= np.linalg.norm(self.line_world_space, axis=1)[:, np.newaxis]
        self.random_numbers = get_random_numbers()
        line_cosine_space = self.line_world_space @ self.world_to_cosine_space.T
        line_length = np.linalg.norm(line_cosine_space[1] - line_cosine_space[0])
        prepared_line = prepare_projected_solid_angle_line_sampling_without_clipping(
            np.asarray([0.0, 0.0, 1.0]), line_cosine_space[0], (line_cosine_space[1] - line_cosine_space[0]) / line_length, line_length)
        self.samples = np.zeros((self.random_numbers.size, 3))
        for j in range(self.random_numbers.shape[0]):
            self.samples[j] = sample_projected_solid_angle_line(prepared_line, self.random_numbers[j])
        self.step_durations = [2.0]
        # Prepare a grid of sphere samples
        self.x_min, self.x_max = -1.05, 1.05
        self.y_min, self.y_max = (-0.35, 1.05)
        dots_per_unit = 200.0
        width, height = (int((self.x_max - self.x_min) * dots_per_unit), int((self.y_max - self.y_min) * dots_per_unit))
        x, y = np.meshgrid(np.linspace(self.x_min, self.x_max, width), np.linspace(self.y_min, self.y_max, height))
        z = np.sqrt(1.0 - x * x - y * y)
        self.view_space_grid = np.stack([x, y, z], axis=2)
        self.world_space_grid = np.einsum("ij, ...j-> ...i", self.projection.matrix.T, self.view_space_grid)

    def on_mouse_move(self, position, motion, held_buttons):
        if "lmb" in held_buttons:
            self.yaw += motion[0] * 0.003
            self.pitch += motion[1] * 0.003
            self.projection = get_euler_projection(self.yaw, self.pitch)
            self.world_space_grid = np.einsum("ij, ...j-> ...i", self.projection.matrix.T, self.view_space_grid)
            self.draw_line()
            return "redraw_canvas"

    def animate(self):
        if self.lerp_ongoing:
            self.draw_line()
            return "redraw_canvas"

    def draw(self):
        draw_title(self.title)
        pyplot.text(8.25, 0.9, "cosine distribution", ha="left", va="bottom")
        if self.step >= 1:
            pyplot.text(8.25, 0.9, "linearly transformed ", ha="right", va="bottom")
        self.slide_axes = pyplot.gca()
        self.line_axes = self.slide_axes.inset_axes((2.0, 1.5, 12.0, 6.0), transform=self.slide_axes.transData, aspect="equal", frameon=False)
        pyplot.gcf().add_axes(self.line_axes)
        self.draw_line()

    def draw_line(self):
        self.line_axes.clear()
        self.line_axes.set_xlim(self.x_min, self.x_max)
        self.line_axes.set_ylim(self.y_min, self.y_max)
        self.line_axes.set_xticks([])
        self.line_axes.set_yticks([])
        pyplot.sca(self.line_axes)
        # Plot the hemisphere
        lerp_factor = np.linspace(0.0, 1.0, 257)
        horizon = np.vstack([np.cos(lerp_factor * 2.0 * np.pi), np.sin(lerp_factor * 2.0 * np.pi), np.zeros_like(lerp_factor)]).T
        self.projection.plot_curve(horizon, True, False, color=horizon_color)
        pyplot.plot(np.cos(lerp_factor * np.pi), np.sin(lerp_factor * np.pi), color=horizon_color)
        # Define the current transform and apply it
        used_world_to_cosine_space = smooth_lerp(self.world_to_cosine_space, self.world_to_shading_space, 1.0 - (self.step_lerp if self.step == 0 else 1.0))
        line_world_space = self.line_world_space @ self.world_to_cosine_space.T
        line_world_space = line_world_space @ np.linalg.inv(used_world_to_cosine_space).T
        line_world_space /= np.linalg.norm(line_world_space, axis=-1)[..., np.newaxis]
        # Plot the transformed horizon
        horizon = horizon @ np.linalg.inv(used_world_to_cosine_space).T
        horizon /= np.linalg.norm(horizon, axis=1)[:, np.newaxis]
        self.projection.plot_curve(horizon, True, False, color=horizon_color)
        # Plot the line outline
        outline = sample_edges(line_world_space, True)
        self.projection.plot_curve(outline, True, False, color=solid_angle_line_color)
        # self.projection.plot_curve(outline, False, True, color=solid_angle_line_color, linestyle="--")
        # Plot the samples
        samples = np.einsum("ij, ...j -> ...i", np.linalg.inv(used_world_to_cosine_space), self.samples)
        samples /= np.linalg.norm(samples, axis=1)[:, np.newaxis]
        samples = self.projection.transform(samples)
        mask = samples[:, 2] > 0.0
        pyplot.scatter(samples[mask, 0], samples[mask, 1], color="w", edgecolors="k", s=12.0, zorder=100.0)
        # Shade the solid angle using rasterization
        cosine_space = np.einsum("ij, ...j-> ...i", used_world_to_cosine_space, self.world_space_grid)
        cosine_space_length = np.linalg.norm(cosine_space, axis=-1)
        ltc_density = np.maximum(0.0, cosine_space[..., 2]) * np.linalg.det(used_world_to_cosine_space) / cosine_space_length ** 4
        ltc_density_rgb = cm.get_cmap("magma")(ltc_density / np.nanmax(ltc_density))[..., 0:3]
        ltc_density_rgb = np.where(np.isnan(ltc_density)[:, :, np.newaxis].repeat(3, 2), 1.0, ltc_density_rgb)
        ltc_density_rgb = np.where((cosine_space[..., 2] < 0.0)[:, :, np.newaxis].repeat(3, 2), 1.0, ltc_density_rgb)
        pyplot.imshow(ltc_density_rgb, extent=(self.x_min, self.x_max, self.y_min, self.y_max), zorder=-1000.0, origin="lower", interpolation="bilinear")


def maximal_error_slide(step):
    """Last slide of the presentation."""
    draw_title("Error analysis")
    draw_bullets(["",
                  "Fix iteration count to 2",
                  r"Worst case error in $\xi$:  $1.6\cdot10^{-5}$",
                  r"$\rightarrow$ unbiased"][0:(3 if step <= 0 else 4)])


class StratificationResultSlide(ImageComparisonSlide):
    """Shows Cornell box renderings."""

    def __init__(self):
        super().__init__()
        self.images = [imread(path.join("../vulkan_renderer/data/experiments",
            [name for name in listdir("../vulkan_renderer/data/experiments") if re.match(r"bistro_inside_%s_.*\.png" % technique, name)][0]))
                for technique in ["independent_blue", "jittered_blue"]]
        self.labels = ["Independent blue noise", "Jittered uniform blue noise"]
        self.left_bottom = 1.0 / 3.0, 1.3
        self.image_width = 7.4
        self.offset = self.image_width + (16.0 - 2.0 * self.image_width) / 3.0, 0.0

    def draw(self):
        super().draw()
        draw_title("Bonus: Stratification")
        # Illustrate independent sampling
        y = self.left_bottom[1] - 0.25
        centers = [self.left_bottom[0] + j * self.offset[0] + self.image_width * 0.5 for j in range(2)]
        random_numbers = np.asarray([0.8, 0.15, 0.3])
        half_line_width = 0.3 * self.image_width
        x = centers[0] + (random_numbers * 2.0 - 1.0) * half_line_width
        pyplot.scatter(x, 3 * [y])
        text_y = y - 0.15
        for j in range(3):
            pyplot.text(x[j], text_y, r"$\xi_%d$" % j, ha="center", va="top")
        # Illustrate jittered uniform sampling
        random_number = 0.7
        x = centers[1] + ((random_number / 3.0 + np.arange(0.0, 0.9, 1.0 / 3.0)) * 2.0 - 1.0) * half_line_width
        pyplot.scatter(x, 3 * [y])
        for j in range(3):
            pyplot.text(x[j], text_y, r"$\frac{\xi+%d}{3}$" % j, ha="center", va="top")
        # Draw lines for both
        for center in centers:
            pyplot.plot([center - half_line_width, center + half_line_width], 2 * [y], color="k", zorder=-100.0)


def timings_slide(step):
    """Shows timings of various sampling techniques."""
    draw_title(r"Timings (RTX 2080 Ti, 1920$\times$1080, %d spp)" % ([1, 128][step]))
    # baseline = 3.25, 2.99
    timings = [
        ("Area", "C1", 0.47, 0.41),
        ("Solid angle,\nours", "C0", 0.75, 0.17),
        # ("Clipped solid angle, ours", "C4", 1.29, 0.17),
        ("Projected,\nours", "C2", 4.40, 2.05),
        ("Projected,\nLi et al.", "C3", 10.0, 5.20),
        ("Projected,\nLi et al.,\nBistro", "C3", 13.6, 10.6),
        ("Projected\nquad, Peters", "C4", 28.9, 11.5),
    ]
    rectangle = np.asarray([(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)])
    width = 2.0
    offset = width + (16.0 - len(timings) * width) / (len(timings) + 1)
    left, bottom = (offset - width), 2.0
    height_scale = 4.2 / (timings[-1][2 + step])
    x = left
    for name, color, many_lights, many_samples in timings:
        height = [many_lights, many_samples][step] * height_scale
        pyplot.fill(x + width * rectangle[:, 0], bottom + height * rectangle[:, 1], color=color)
        pyplot.text(x + 0.5 * width, bottom - 0.2, name, ha="center", va="top")
        if step == 0:
            time_text = "%.0f $\mu$s" % (1000.0 * many_lights / 128.0)
        else:
            time_text = "%.0f $\mu$s" % (1000.0 * many_samples)
        pyplot.text(x + 0.5 * width, bottom + height + 0.2, time_text, ha="center", va="bottom")
        x += offset


def conclusions_slide(step):
    draw_title("Conclusions")
    draw_bullets([
        "Linear lights are cheap",
        "Blue noise works great",
        "Newton is bad for sampling",
        "Put it into your renderer now",
        "Code at https://momentsingraphics.de/HPG2021.html",
        "Thanks!"
    ][0:[1, 2, 3, 5, 6][step]], 1.0, 6.5)


if __name__ == "__main__":
    pyplot.style.use("../Data/slides.mplstyle")
    talk = "long"
    presentation = PresentationWithFooter("../Graphics/kit_logo/kit_logo.png", 128.0 * 16.0 / 1920.0, [0, 1] if talk != "fast" else [0, 2])
    if talk == "long":
        presentation.slides += [
            FunctionSlide(title_slide),
            GameSlide(),
            RendererSlide(2, False, True, ""),
            CylinderSlide("linear_light"),
            GamitoSlide(),
            MotivationSlide("area"),
            MotivationResultSlide(1),
            MotivationSlide("solid"),
            MotivationResultSlide(2),
            MotivationSlide("projected"),
            MotivationResultSlide(3),
            TzuMaoLiSlide(),
            CylinderSlide("solid_angle"),
            CylinderSlide("projected_solid_angle"),
            CDFSlide("s"),
            CDFSlide("angle"),
            FunctionSlide(maximal_error_slide, 2),
            CDFSlide("newton"),
            LTCResultsSlide(False),
            MotivationSlide("ltc"),
            LTCSlide("Linearly transformed cosines [Heitz et al. 2016]"),
            LTCResultsSlide(True),
            FunctionSlide(timings_slide, 2),
            FunctionSlide(conclusions_slide, 5),
            StratificationResultSlide(),
        ]
    elif talk == "fast":
        presentation.slides += [
            RendererSlide(2, False, True, ""),
            LTCSlide("Diffuse and specular importance sampling"),
        ]
    # export_presentation_pdf(presentation, "../Data/hpg_slides.pdf", False)
    slideshow = Slideshow(presentation, 0, 0)
    slideshow.start()
